from cast_common.logger import Logger
from cast_common.util import create_folder, yes_no_input,folder_input,secret_input,url_input,string_input
from os import mkdir,walk,listdir
from os.path import exists,abspath,isdir
from glob import glob
from shutil import copytree,rmtree
from oneclick.discovery.sourceValidation import SourceValidation 
from oneclick.configTest import Config,Status
from oneclick.exceptions import InvalidConfiguration


from tqdm import tqdm

class Prepare(SourceValidation):

    @property
    def choose(self) -> bool:
        return True
    
    @property
    def name(self) -> str:
        return __class__.__name__

    def __init__(self):
        pass


    def add_app(self,app_name:str):
        config = self.config
        
        print ("When running an MRI analysis include:")
        security = yes_no_input('\tSecurity analysis', default_value=True)
        blueprint = yes_no_input('\tBlueprint analysis', default_value=True)
        if yes_no_input(f'Add {app_name}?', default_value=True):
            config.log.info(f'Adding {app_name}')
            config.add_app(app_name, security, blueprint)
            create_folder(abspath(f'{config.deliver_folder}/{config.project_name}/{app_name}'))
        pass

    def run(self):
        config = self.config
        self.log.info(f'Starting project: {config.project_name}')

        print(self.show_progress())

        #scan delivery folder for application folders
        dir=[]
        project_folder = abspath(f'{config.deliver_folder}\\{config.project_name}')
        dirnames = listdir(project_folder) 
        if len(dirnames) == 0:
            config.log.warning(f'No applications found in {project_folder}')
            while True:
                if len(config.applist) > 0:
                    # for i in range(7):
                    #     print('\033[A',end='')
                    if not yes_no_input('Add another application?', default_value=False):
                        break
                self.add_app(string_input('What is the name of the application'))
                pass
            pass
        else:
            pass

        dirnames = listdir(project_folder) 
        self.log.info(f'Found {len(dirnames)} applications in {project_folder}')
        for folder in dirnames:
            app_folder = abspath(f'{project_folder}\\{folder}')

            found = False
            for app in config.applist:
                if app['name'] == folder:
                    found = True
                    break
            if not found:
                if yes_no_input (f'Found an extra folder, "{app_folder}", that does not exist in your project, add it now'):
                    self.add_app(folder)
            else:
                self.log.debug(f'{app_folder} found in the project configuration file')

            # self.log.info(f'Looking for source code in {app_folder}')
            files = []
            while len(files) == 0:
                files = glob(f'{app_folder}\\*')
                if len(files) == 0:
                    if not yes_no_input(f'{app_folder} has no source files, please add it now?',default_value=True):
                        self.log.info('Terminated by user')
                        exit(1)
            print(self.show_progress())
            pass
        
        pass
        
        self.log.info(config.application_report())

        for app in config.applist:
            app_folder = abspath(f'{project_folder}\\{app["name"]}')
            stage_folder = abspath(f'{config.stage_folder}\\{config.project_name}\\{app["name"]}')
            if Status(app['status']['aip']) < Status.STAGED:
                if glob(f'{stage_folder}\\*'):
                    self.log.info(f'Removing {stage_folder}')
                    rmtree(stage_folder)
                create_folder(stage_folder)
                self.log.info(f'Staging {app_folder} to {stage_folder}')
                status = copytree(app_folder,stage_folder,dirs_exist_ok=True)
                self.log.info(f'All files copied to {status}')
                app['status']['aip']=app['status']['highlight']=Status.STAGED
                config._save()
            else:
                self.log.info(f'{app["name"]} already staged')

            print(self.show_progress())
            pass

        self.log.info(f'Environment preparation complete for project {config.project_name}')

        return True




"""
    Read and validate configuration file
"""
from cast_common.logger import Logger,DEBUG, INFO, WARN, ERROR
from cast_common.util import create_folder
from json import load
from argparse import ArgumentParser
from json import JSONDecodeError,dump
from os.path import abspath,exists,dirname
from os import getcwd

from cast_common.util import yes_no_input,folder_input,secret_input,url_input,string_input

from argparse import ArgumentParser
from oneclick.exceptions import NoConfigFound,InvalidConfiguration,InvalidConfigNoBase

__author__ = "Nevin Kaplan"
__copyright__ = "Copyright 2022, CAST Software"
__email__ = "n.kaplan@castsoftware.com"

class Config():
    log = None
    log_translate = {} 

    def __init__(self,parser:ArgumentParser,default_args={},log_level: int=INFO):
        self.log = Logger(self.__class__.__name__,log_level)

        args = parser.parse_args()
 
        
        if 'baseFolder' not in args:
            args.baseFolder = folder_input('What folder is OneClick installed in?',file='.oneClick',use_default=True)

        if not exists(abspath(args.baseFolder)) and exists(abspath(f'args.baseFolder/.oneClick')):
            raise InvalidConfigNoBase(f'Invalid base folder: {args.baseFolder}')

        base_config=abspath(f'{args.baseFolder}/.oneclick/config.json'.lstrip().rstrip())
        if not exists(base_config) and 'command' in args and args.command == 'run':
            raise NoConfigFound('Base configuration file found, please run with the "config" option')


        if 'command' not in args or args.command == 'config':
            if not exists(base_config):
                self._config={}
                self._config_file=base_config
                self.base = args.baseFolder.lstrip().rstrip()
            else: 
                if 'projectName' not in args:
                    self._config_file = base_config
                else:
                    self._config_file = abspath(f'{args.baseFolder}/.oneclick/{args.projectName}.json'.lstrip().rstrip())

                with open(abspath(self._config_file), 'rb') as config_file:
                    self._config = load(config_file)

            #if there is no command in arguments we are comming from setup and should exit here
            if 'command' not in args:
                return
            
            try:     
                if 'java_home' not in args: 
                    self.java_home = ''
                else:
                    self.java_home = args.java_home

                if 'cloc_version' not in args: self.cloc_version = args.cloc_version
                if 'profiler' not in args is not None: self.profiler = args.profiler

                #Dashboard
                if args.aipURL is not None: self.aip_url = args.aipURL
                if args.aipPassword is not None: self.aip_password = args.aipPassword
                if args.aipUser is not None: self.aip_user = args.aipUser

                #Highlight
                if self.check_default(args.hlURL,self.hl_url,default_args['hlURL']):
                    self.hl_url = f'{args.hlURL}/WS2'
                if args.hlUser is not None: self.hl_user = args.hlUser
                if args.hlPassword is not None: self.hl_password = args.hlPassword
                if args.hlInstance is not None: self.hl_instance = args.hlInstance
                if self.check_default(args.hlCLI,self.hl_cli,default_args['hlCLI']):
                    self.hl_cli = args.hlCLI

                self.perl_install_dir = abspath(f'{args.hlAgent}/strawberry/perl')
                self.analyzer_dir = abspath(f'{args.hlAgent}/perl')

                #AIPConsole
                if args.consoleURL is not None: self.console_url = args.consoleURL
                if args.consoleKey is not None: self.console_key = args.consoleKey
                if args.consoleCLI is not None: self.console_cli = args.consoleCLI
                if args.enable_security_assessment is not None: self.enable_security_assessment = args.enable_security_assessment
                if args.blueprint is not None: self.blueprint = args.blueprint

                #Database
                if args.dbHost is not None: self.db_host = args.dbHost
                if args.dbPort is not None: self.db_port = args.dbPort
                if self.check_default(args.dbUser,self.db_user,default_args['dbUser']):
                    self.db_user = args.dbUser
                if self.check_default(args.dbPassword,self.db_password,default_args['dbPassword']):
                    self.db_password = args.dbPassword
                if self.check_default(args.dbDatabase,self.db_database,default_args['dbDatabase']):
                    self.db_database = args.dbDatabase

                #settings
                if self.check_default(args.cloc_version,self.cloc_version,default_args['cloc_version']):
                    self.cloc_version = args.cloc_version
                if self.check_default(args.profiler,self.profiler,default_args['profiler']):
                    self.profiler = args.profiler
        
            except AttributeError as e:
                self.log.error(str(e))
            return

        #do all required fields contain data
        if args.command == 'run':
            try:
                self._config={}
                self._config_file = abspath(f'{args.baseFolder}/.oneclick/{args.projectName}.json')
                if exists(self._config_file):
                    with open(abspath(self._config_file), 'rb') as config_file:
                        self._config = load(config_file)
                else:
                    with open(base_config) as config_file:
                        self._config = load(config_file)

                self.base=args.baseFolder
                self.project = args.projectName
                self.company_name = args.companyName

                self.start=args.start
                self.end=args.end

                # Run for MRI
                if self.is_console_active == False:
                    if not args.runMRI is None:
                        self._set_aip_active=args.runMRI
                    elif yes_no_input('Run MRI analysis for all applications?'):
                        while not self.is_console_active:
                            if len(self.console_url) == 0:
                                self.console_url=url_input('Missing console URL:',self.console_url)
                            else:
                                self.console_url=self.console_url
                            if len(self.console_key) == 0:
                                self.console_key=secret_input('Missing console KEY:',self.console_key)
                            if len(self.console_cli) == 0:
                                self.console_cli=folder_input('\t"AIP Console automation tools" location',dirname(self.console_cli),"aip-console-tools-cli.jar",True)
                    else:
                        self._set_console_active=False                                

                if self.is_hl_active == False:                
                    if not args.runHL is None:
                        self._set_hl_active=args.runHL
                    elif yes_no_input('Run Highlight analysis for all applications?'):
                        while not self.is_hl_active:
                            if len(self.hl_url) == 0:
                                self.hl_url=url_input('Missing Highlight URL',self.hl_url)
                            if len(self.hl_user) == 0:
                                self.hl_user=string_input('Missing Highlight User ID',self.hl_user)
                            if len(self.hl_password) == 0:
                                self.hl_password=secret_input('Missing Highlight Password',self.hl_password)
                            if len(self.hl_instance) == 0:
                                self.hl_instance=string_input('Missing Highlight Instance ID',self.hl_instance)
                            else:
                                self.hl_instance=self.hl_instance
                    else:
                        self._set_hl_active=False                                

                # self.log.info(f'Run MRI analysis: {self.is_aip_active}')
                # self.log.info(f'Run Highlight analysis: {self.is_hl_active}')
                self._save()

            except JSONDecodeError as e:
                msg = str(e)
                self.log.error(f'Configuration file {self._config_file} must be in a JSON format {msg}')
                exit()

            except ValueError as e:
                msg = str(e)
                self.log.error(msg)
                exit()

    def validate_for_run(self):
        if self.java_home == '':
            raise InvalidConfiguration('Missing JAVA HOME location')

        if self.profiler == '':
            raise InvalidConfiguration('Missing profiler executable location')
        exec = abspath(self.profiler)
        if not exists(exec):
            raise InvalidConfiguration(f'Profiler executable not found: {exec}')

        if self.cloc_version == '':
            raise InvalidConfiguration('Missing CLOC executable location')
        exec = abspath(f'{self.base}\\scripts\\{self.cloc_version}')
        if not exists(exec):
            raise InvalidConfiguration(f'CLOC executable not found: {exec}')
        # if not self.is_console_active and not self.is_hl_active:
        #     raise InvalidConfiguration('Both Hightlight and AIP configureations are incomplete, at least one must be')


    def check_default(self,arg_value,cfg_value,default_value) -> bool:
        rtn =  False
        if arg_value is not None: 
            if (cfg_value is None or cfg_value=='') or arg_value != default_value:
                rtn =  True
        return rtn


    def clean_creds(self,value):
        if type(value) is dict:
            for skey, svalue in value.items():  
                if skey=='password' or skey=='user':
                    value[skey]='*******************'
                else:
                    self.clean_creds(svalue)
        return

    def _if_set(self,target,param):
        if param is not None:
            target[0]=param

    def _save(self):
        create_folder(abspath(f'{self.base}/.oneClick'))
        with open(self._config_file, "w") as f:
            dump(self._config, f,indent=4)

    def _set_rest_settings(self,dict):
        for v in ["Active","URL","user","password"]:
            if v not in dict:
                raise ValueError(f"Required field '{v}' is missing from config.json")
    @property
    def rest(self):
        return self._get(self._config,'rest',{})

    def _set_active(self,node,chk_lst):
        chk = True
        for item in chk_lst:
            if item not in node or node[item]=='':
                chk = False
                break
        node['Active']=chk

    def _get(self,base,key,default=''):
        if key not in base:
            base[key] = default  
        return base[key]        

    def _set_value(self,base,key,value,default=''):
        base[key] = self._get(base,key,default)
        if value is not None:
            base[key]=value
            return True
        else:
            return False

    """ **************** Highlight entries ************************ """
    @property
    def highlight(self):
        if 'Highlight' not in self.rest:
            self.rest['Highlight']={}
            self.highlight['Active']=False
        return self.rest['Highlight']
    @property
    def is_hl_active(self):
        return self.highlight['Active']

    def _set_hl_active(self):
        self._set_active(self.highlight,['URL','cli','perlInstallDir','analyzerDir','user','password','instance'])

    @property
    def hl_url(self):
        return self._get(self.highlight,'URL')
    @hl_url.setter
    def hl_url(self,value):
        if self._set_value(self.highlight,'URL',value):
            self._set_hl_active()
            self._save()

    @property
    def hl_user(self):
        return self._get(self.highlight,'user')
    @hl_user.setter
    def hl_user(self,value):
        if self._set_value(self.highlight,'user',value):
            self._set_hl_active()
            self._save()

    @property
    def hl_password(self):
        return self._get(self.highlight,'password')
    @hl_password.setter
    def hl_password(self,value):
        if self._set_value(self.highlight,'password',value):
            self._set_hl_active()
            self._save()

    @property
    def hl_instance(self):
        return self._get(self.highlight,'instance')
    @hl_instance.setter
    def hl_instance(self,value):
        if self._set_value(self.highlight,'instance',value):
            self._set_hl_active()
            self._save()

    @property
    def hl_cli(self):
        return self._get(self.highlight,'cli')
    @hl_cli.setter
    def hl_cli(self,value):
        if self._set_value(self.highlight,'cli',value):
            self._set_hl_active()
            self._save()

    @property
    def perl_install_dir(self):
        return self._get(self.highlight,'perlInstallDir')
    @perl_install_dir.setter
    def perl_install_dir(self,value):
        if self._set_value(self.highlight,'perlInstallDir',value):
            self._set_hl_active()
            self._save()

    @property
    def analyzer_dir(self):
        return self._get(self.highlight,'analyzerDir')
    @analyzer_dir.setter
    def analyzer_dir(self,value):
        if self._set_value(self.highlight,'analyzerDir',value):
            self._set_hl_active()
            self._save()

    @property
    def is_hl_config_valid(self)->bool:
        if not self.is_hl_active:
            self.log.warning('Highlight analysis is inactive')            
            return True

        is_ok=True
        msg=[]
        if len(self.hl_cli) == 0:
            msg.append('hl_cli')
        if len(self.perlInstallDir) == 0:
            msg.append('perlInstallDir')
        if len(self.analyzerDir) == 0:
            msg.append('analyzerDir')
       
        if len(msg) > 0:
            fmt_msg=', '.join(msg)
            self.log.error(f'Invalid Highlight configuration, missing {fmt_msg} fields')
            is_ok=False

        return is_ok

    """ **************** AIP REST related entries ************************ """
    def _set_aip_active(self,key,value,default=''):
        if self._set_value(self.aip,key,value,default):
            self._set_active(self.aip,['URL','user','password'])
        self._save()

    @property
    def aip(self):
        return self._get(self.rest,'AIP',{})

    @property
    def is_aip_active(self)->bool:
        return self._get(self.aip,'Active',False)

    @property
    def aip_url(self):
        return self._get(self.aip,'URL')
    @aip_url.setter
    def aip_url(self,value):
        self._set_aip_active('URL',value,'')
            
    @property
    def aip_user(self):
        return self._get(self.aip,'user')
    @aip_user.setter
    def aip_user(self,value):
        self._set_aip_active('user',value,'')

    @property
    def aip_password(self):
        return self._get(self.aip,'password')
    @aip_password.setter
    def aip_password(self,value):
        self._set_aip_active('password',value,'')

    """ **************** Console REST related entries ************************ """
    def _set_console_active(self,key,value,default=''):
        if self._set_value(self.console,key,value,default):
            self._set_active(self.console,['URL','API_Key','cli'])
            self._save()

    @property
    def is_console_active(self)->bool:
        return self._get(self.console,'Active',False)

    @property
    def console(self):
        return self._get(self.rest,'AIPConsole',{})

    @property
    def console_url(self)->str:
        return self._get(self.console,'URL')
    @console_url.setter
    def console_url(self,value):
        self._set_console_active('URL',value,'')

    @property
    def console_key(self)->str:
        return self._get(self.console,'API_Key')
    @console_key.setter
    def console_key(self,value):
        self._set_console_active('API_Key',value,'')

    @property
    def console_cli(self):
        return self._get(self.console,'cli')
    @console_cli.setter
    def console_cli(self,value):
        self._set_console_active('cli',value,'')

    @property
    def console_node(self):
        return self._get(self.console,'node')
    @console_node.setter
    def console_node(self,value):
        self._set_console_active('node',value,'')

    @property
    def enable_security_assessment(self):
        return self._get(self.console,'enable-security-assessment')
    @enable_security_assessment.setter
    def enable_security_assessment(self,value):
        self._set_console_active('enable-security-assessment',value,'')

    @property
    def blueprint(self):
        return self._get(self.console,'blueprint')
    @blueprint.setter
    def blueprint(self,value):
        self._set_console_active('blueprint',value,'')

    @property
    def is_console_config_valid(self)->bool:
        if not self.is_console_active:
            self.log.warning('MRI analysis is inactive')            
            return True
        is_ok=True
        msg=[]
        if len(self.console_url) == 0:
            msg.append('URL')
        if len(self.console_key) == 0:
            msg.append('API_Key')
        if len(self.console_cli) == 0:
            msg.append('Console CLI')
       
        if len(msg) > 0:
            fmt_msg=', '.join(msg)
            self.log.error(f'Invalid MRI configuration, missing {fmt_msg} fields')
            is_ok=False

        return is_ok


    """ **************** Action Plan related entries ************************ """
    def _set_database_active(self,key,value,default=''):
        if self._set_value(self.db,key,value,default):
            self._set_active(self.db,['database','user','password','host','port'])
            self._save()

    @property
    def is_db_active(self)->bool:
        return self.db['Active']

    @property
    def db(self):
        return self._get(self._config,'Database',{})

    @property
    def db_database(self):
        return self._get(self.db,'database')
    @db_database.setter
    def db_database(self,value):
        self._set_database_active('database',value,'')

    @property
    def db_user(self):
        return self._get(self.db,'user')
    @db_user.setter
    def db_user(self,value):
        self._set_database_active('user',value,'')

    @property
    def db_password(self):
        return self._get(self.db,'password')
    @db_password.setter
    def db_password(self,value):
        self._set_database_active('password',value,'')
    
    @property
    def db_host(self):
        return self._get(self.db,'host')
    @db_host.setter
    def db_host(self,value):
        self._set_database_active('host',value,'')
    
    @property
    def db_port(self):
        return self._get(self.db,'port')
    @db_port.setter
    def db_port(self,value):
        self._set_database_active('port',value,'')



    """ **************** Project related entries ************************ """
    @property
    def project(self):
        return self._config['project']
    @project.setter
    def project(self,value):
        if type(value) is not str:
            raise ValueError(f'Expecting a the project name, got {type(value)}')

        self.log.info(f'Starting executin of project {value}')
        if 'project' not in self._config:
            self._config['project']={}
            self._config['project']['name']=value
            self._config['project']['application']={}

        elif value.lower() != self.project_name.lower():
            raise ValueError("Can't rename a project")

    @property
    def project_name(self):
        return self.project['name']

    @property
    def company_name(self):
        return self.project['company_name']
    @company_name.setter
    def company_name(self,value):
        if value is not None:
            self.project['company_name']=value
            self._save()

    """ **************** Application related entries ************************ """
    @property
    def application(self):
        return self.project['application']
    @application.setter
    def application(self,value:list):
        if type(value) is not list:
            raise ValueError(f'Expecting a list of application names, got {type(value)}')

        update = False
        for app in list(self.application):
            if app not in value:
                del self.application[app]
                update = True

        for appl_name in value:
            if appl_name in self.application.keys():
                pass
            else:
                self.project['application'][appl_name]={'aip':'','hl':''}
                update = True
        


        # self._config['application']=value
        if update:
            self._save()

    @application.deleter
    def application(self):
        self.project['application']={}

    @property
    def deliver(self):
        return f'{self.workbase}\\deliver\\{self.project_name}'
    @property
    def work(self):
        return f'{self.workbase}\\STAGED'
    @property
    def report(self):
        return f'{self.workbase}\\REPORT'
    @property
    def logs(self):
        return f'{self.workbase}\\LOGS'
    @property
    def oneclick_work(self):
        return f'{self.workbase}\\ONECLICK_WORK'

    """ **************** Email related entries ************************ """
    @property 
    def email(self):
        return self._config['email']
    @property
    def from_email_addrs(self):
        return self.email['from']

    @property
    def from_email_passwd(self):
        return self.email['password']

    @property
    def to_email_addrs(self):
        return self.email['to']

    @property
    def email_body(self):
        return self.email['body']     

    @property
    def email_subject(self):
        return self.email['subject'] 

    """ **************** Setting related entries ************************ """
    @property 
    def setting(self):
        return self._get(self._config,'setting',{})        

    @property
    def arg_template(self):
        return self._get(self.setting,'arg-template')

    @property
    def workbase(self):
        return self._get(self.setting,'workbase','')
    @workbase.setter
    def workbase(self,value):
        if value is not None:
            self.setting['workbase']=value
            self._save()

    @property
    def base(self):
        return self._get(self.setting,'base','')
    @base.setter
    def base(self,value):
        if value is not None:
            self.setting['base']=value
            self._save()

    @property
    def profiler(self):
        return self._get(self.setting,'profiler')
    @profiler.setter
    def profiler(self,value):
        if len(value):
            value = abspath(value)
        if self._set_value(self.setting,'profiler',value):
            self._save()

    @property
    def cloc_version(self):
        return self._get(self.setting,'cloc_version')
    @cloc_version.setter
    def cloc_version(self,value):
        if self._set_value(self.setting,'cloc_version',value):
            self._save()

    # @property
    # def reset(self):
    #     return self.setting['reset']
    # @reset.setter
    # def reset(self,value):
    #     if value is None:
    #         self.setting['reset']=False
    #     else:
    #         self.setting['reset']=value
    #     self._save()

    @property
    def java_home(self):
        if 'java-home' not in self.setting: self.setting['java-home']=''
        return self.setting['java-home']
    @java_home.setter
    def java_home(self,value):
        if self._set_value(self.setting,'java-home',value,''):
            self._save()

#        def _set_value(self,base,key,value,default=''):


    
    # @property
    # def template(self):
    #     return self._config['template']


from cast_common.logger import Logger,INFO
from oneclick.config import Config
from oneclick.analysis.analysis import Analysis,Process
from subprocess import TimeoutExpired
from time import sleep
#from util import find_in_list

class TrackAnalysis(Analysis):
    @property
    def name(self) -> str:
        return f'Track Analysis'
    def choose(self) -> bool:
        return True

    def __init__(cls):
        pass

    def run(cls):
        config = cls.config

        print(cls.show_progress(clear=True))
        config.log.warning('Stopping oneclick will stop HL analysis but not AIP.')
        config.log.info(f'The status of AIP ananlsys can be tracked directly on AIP Console at {config.console_url}')
        print(cls.show_progress())

        while True:
            # cls.log.info(f'Checking {len(cls._pid)} processes ...')
            for p in cls._pid:
                just_completed=False
                process = p.process
                if p.status is None or p.status == "Running":
                    if process is not None and process.poll() is None:
                        #process is running
                        p.status = "Running"
                        try:
                            # line = process.stdout.readline(timeout=1)
                            # line = line.rstrip('\n')
                            # if len(line.strip(' ')) > 0:
                            #     p.log.append(line)

                            stdout, stderr = process.communicate(timeout=1)
                            for line in stdout.split('\n'):
                                p.log.append(line)

                        except TimeoutExpired:
                            pass 
                        pass
                    else:
                        just_completed=True
                        if process is None:
                            if p.operation == 'AIP':
                                p.status = config.application[p.name]['aip']
                            else:
                                p.status = config.application[p.name]['hl']
                        else:
                            stdout, stderr = process.communicate()
                            for line in stdout.split('\n'):
                                #if len(find_in_list(line,p.log))==0:
                                p.log.append(line)
                            if process.returncode == 0:
                                p.status = 'Complete'
                            else:
                                p.status = f'Error: {process.returncode}'
                                cls.log.error(f'error running {p.name}')
                                cls.log.error('\n'.join(p.log))

                            if p.operation == 'AIP':
                                config.application[p.name]['aip'] = p.status
                            else:
                                config.application[p.name]['hl'] = p.status
                            config._save()
                            

                cls.log.info(f"{p.operation} analysis for Application {p.name}, {p.status}")
                if just_completed == True:
                    if p.status != 'Complete':
                        for line in p.log:
                            print(f'\t{line}')

                    if p.status=='Complete' and  p.operation == 'AIP':
                        # cls.log.info('Running post analsis jobs...')
                        for proc in cls._post_aip:
                            if proc.__class__.__name__ not in config.application[p.name] or \
                                config.application[p.name][proc.__class__.__name__] != 'Complete':

                                cls.log.info(f'******************* {proc.__class__.__name__} *******************************')
                                proc.run(p.name)
                                
                                config.application[p.name][proc.__class__.__name__]='Complete'
                                config._save()


            running = False
            for p in cls._pid:
                if p.status=='Running':
                    running = True
                    break 
            if not running:
                cls.log.info(f'All processing complete')
                error = False
                for p in cls._pid:
                    cls.log.info(f"{p.operation} for {config.project_name}\{p.name}: {p.status}")
                    if "Complete" not in p.status:
                        error = True
                break
            sleep(60)
        return error


        #     status,output = check_process(process[appl])
        #     if status != 0:
        #         cls.log.error(f'Error analyzing {appl}')
        #         error = True
        # if error:
        #     # TODO: add more desriptive error message
        #     raise RuntimeError ("")

    def get_title(cls) -> str:
        return "TRACKING ANALYSIS" 



from oneclick.analysis.analysis import Analysis
from oneclick.analysis.trackAnalysis import TrackAnalysis
from oneclick.config import Config
from cast_common.logger import INFO
from cast_common.util import run_process
from json import dumps
from os.path import abspath


class AIPAnalysis(Analysis):
    @property
    def choose(self) -> bool:
        return True
    @property
    def can_run(cls) -> bool:
        return cls.config.console==True

    @property
    def name(self) -> str:
        return f'MRI Analysis'
    def choose(self) -> bool:
        return True

    def __init__(cls):
        cls._df = {}
        pass

    def __init__(cls):
        pass
    
    def run(cls):
        config = cls.config
        if not config.console:
            cls.config.log.warning('AIP Console is not configured, analysis will not run')
            return 0

        for appl in config.application:

            #has thi spplication already been run?
            aip_status = config.application[appl]['aip']
            if aip_status == '' or aip_status.startswith('Error'):
                #add a new appication in AIP Console
                # cls.log.info(f'Running analysis for {config.project_name}\{appl}')
                cls.log.info(f'Running AIP analysis for {appl}')

                java_home = config.java_home
                if len(java_home) > 0:
                    java_home = f'{java_home}'

                node_name = ""
                if len(config.console_node) > 0:
                    node_name=f'--node-name={config.console_node}'

                security_assessment=""
                if config.enable_security_assessment:
                    security_assessment='--enable-security-assessment'
                
                blueprint=""
                if config.blueprint:
                    blueprint='--blueprint'

                args = [abspath(f'{java_home}/bin/java.exe'),
                        '-jar',config.console_cli,
                        'add',
                        '-n',appl,
                        '-f', f'AIP/{config.project_name}/{appl}',
                        '-s',config.console_url,
                        '--apikey',config.console_key,
                        '--verbose',
                        '--auto-create',
                        node_name, security_assessment, blueprint
                        ]
                cls.log.debug(dumps(args, indent=2))

                try:
                    process = run_process(args,wait=False)
                except FileNotFoundError as e:
                    cls.log.error(f'Unable to launch analysis process {e}')
                    cls.log.error(args)
                    return e.errno
            else:
                cls.log.info(f'{appl} has already been successfully analyized, skipping step')
                process = None

            cls.track_process(process,"AIP",appl)
        return 0

        TrackAnalysis(INFO).run(config)
        
        


        


__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2024, CAST Software"

from oneclick.configTest import Config,Status
from cast_common.logger import Logger,DEBUG, INFO, WARN, ERROR
from cast_common.util import create_folder
from oneclick.exceptions import InvalidConfigNoBase


from os.path import abspath,join

class Base():
    __config = None
    __log = None
    _progress = False

    def __init__(self, config:Config=None):
        if config is None and Base.__config:
            raise InvalidConfigNoBase()
        elif config is not None:
            Base.__config = config

        if self.log is None:       
            log_level = config.log_level
            log_folder = abspath(f'{config.base_log_folder}/{config.project_name}')
            create_folder(log_folder)
            self.log = Logger(self.__class__.__name__,log_level,file_name=abspath(f'{log_folder}/General.log'),console_output=False)

    @property
    def config(self):
        return Base.__config

    @property
    def log(self):
        return Base.__log    
    @log.setter
    def log(self, value):
        Base.__log = value

    @property
    def choose(self) -> bool:
        return True if self.config.debug else False
    
    @property
    def can_run(self) -> bool:
        return True
    
    process_cnt = 0
    process_txt = '\\|/-\\|/-'
    def show_progress(self,done=False,clear=False):
        config = self.config

        out = []

        header = f'------------------------ ------------------------ ---------- -------------- --------- --------- --------- --------- --------- --------- ---------'
        if clear:
            if Base._progress:
                for l in range(len(config.applist)+5):
                    line = ' '
                    out.append(f'{line:{len(header)}}\n')
        else:
            Base._progress=True
            line = f'{self.process_txt[self.process_cnt]} Running {self.name} for project: {config.project_name}'
            out.append(f'\r{line:<{len(header)}}\n')
            out.append(f'                                                                                  Removed                            SQL Discovery\n')
            out.append(f'Appl Name                Status                    Unpacked       LOC        Folders     Files   Tables   Functions   Procs     Views   Triggers\n')
            out.append(f'{header}\n')
            for app in config.applist:
                app_name = app['name']
                app_status = app['status']['aip']
                loc = app['loc'] if 'loc' in app else ''
                unpacked = app['unpacked'] if 'unpacked' in app else ''
                folders = app['deleted']['folders'] if 'deleted' in app else ''
                files = app['deleted']['files'] if 'deleted' in app else ''
                tables = app['sql']['tables'] if 'sql' in app else ''
                functions = app['sql']['functions'] if 'sql' in app else ''
                procedures = app['sql']['procedures'] if 'sql' in app else ''
                views = app['sql']['views'] if 'sql' in app else ''
                triggers = app['sql']['triggers'] if 'sql' in app else ''
                
                out.append(f'{app_name:<25}{Status(app_status).name.replace("_", " "):<25}{unpacked:>10}{loc:>15}{folders:>10}{files:>10}{tables:>10}{functions:>10}{procedures:>10}{views:>10}{triggers:>10}\n')

        if done:
            return ''.join(out)
        elif Base._progress:
            pass
            for l in range(len(config.applist)+5):
                out.append(f'\033[F')
            out.append('\r')
            self.process_cnt += 1
            if self.process_cnt == len(self.process_txt):
                self.process_cnt = 0
            out.append(f'\r\033[?25l')
            return ''.join(out)
        else:
            return ''.join(out)        



from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.mime.text import MIMEText
from string import Template

from oneclick.discovery.sourceValidation import SourceValidation
from os.path import abspath
from oneclick.config import Config

import smtplib

class EmailNotification(SourceValidation):

    def __init__(cls, config:Config, log_level:int):
        super().__init__(config,cls.__class__.__name__,log_level)
        pass

    def run(cls, config:Config):

        # set up the SMTP server
        s = smtplib.SMTP(host='smtp.office365.com', port=587)
        s.starttls()

        s.login(config.from_email_addrs, config.from_email_passwd)

        # For each contact, send the email:
        for email in list(config.to_email_addrs):
            msg = MIMEMultipart()       # create a message

            # setup the parameters of the message
            msg['From']=config.from_email_addrs
            msg['To']=email
            msg['Subject']=config.email_subject
            message = config.email_body
            # Prints out the message body for our sake
            # print(message)
            cls._log.info(f'sending a mail to {email}')
            # add in the message body
            msg.attach(MIMEText(message, 'plain'))

            discovery_report = abspath(f'{config.report}/{config.project_name}/{config.project_name}-source-code-discovery.docx')
            # attaching docx file
            docx = MIMEApplication(open(discovery_report, 'rb').read())
            docx.add_header('Content-Disposition', 'attachment', filename= f'{config.project_name}-source-code-discovery.docx')
            msg.attach(docx)
            
            # send the message via the server set up earlier.
            s.send_message(msg)
            cls._log.info(f'sent a mail containing discovery report to {email}')
            #del msg
            
        # Terminate the SMTP session and close the connection
        s.quit()
    




from datetime import datetime
from os import mkdir,getcwd,walk,remove,chmod,rmdir
from stat import S_IWUSR
from os.path import abspath,join
#from shutil import rmtree
from oneclick.discovery.sourceValidation import SourceValidation
from oneclick.configTest import Config,Status
from cast_common.logger import Logger,INFO
from cast_common.util import create_folder

from pandas import Series,DataFrame

import re 

#todo: review cleanup lists for aip and hl, do we need separate or can we keep it as one and run HL from AIP folder?

class Cleanup(SourceValidation):
    @property
    def choose(self) -> bool:
        return False
    
    @property
    def name(self) -> str:
        return 'Cleanup'

    def __init__(cls):
        pass
    
    @property
    def cleanup_file_prefix(cls):
        return ""

    def run(cls):
        config = cls.config
        config.log.debug('Source Code cleanup in progress')

        #have the minimum requirements been met to run cleanup
        can_run = True
        for appl in config.applist:
            if appl['status']['aip'] < Status.CLOC_PRE_CLEAN_END:
                can_run = False
                break
        if not can_run:
            config.log.info('Cleanup cannot run, please run "Technology Discovery - Before Cleanup" step first')
            return False

        # for app in config.applist:
        #     app_status = app['status']
        #     if app_status['aip'] < Status.SOURCE_CLEAN_START:
        #         app['deleted'] = {'folders': 0, 'files': 0}


        output_path = abspath(f'{config.log_folder}/{config.project_name}')

        dateTimeObj=datetime.now()
        file_suffix=dateTimeObj.strftime("%d-%b-%Y(%H.%M.%S.%f)")
        
        exclusionFileList= abspath(f'{config.base}/scripts/{cls.cleanup_file_prefix}deleteFileList.txt')
        with open(exclusionFileList) as f:
            files_list = f.read().splitlines()

        exclusionFolderList= abspath(f'{config.base}/scripts/{cls.cleanup_file_prefix}deleteFolderList.txt')
        with open(exclusionFolderList) as f:
            folder_list = f.read().splitlines()

        apps = config.applist
        # cls._log.info(f'Running {cls.__class__.__name__} for all applications')

        for app in apps:
            app_status = app['status']
            if app_status['aip'] < Status.UNPACK_END:
                config.log.info(f'Cleanup cannot run, please run "Source Code Discovery - Unpack" step first')
                return False
            
            if app_status['aip'] < Status.SOURCE_CLEAN_START:
                app_status['aip'] = app_status['highlight'] = Status.SOURCE_CLEAN_START

        print(cls.show_progress())

        for app in apps:
            app_name = app['name']
            if 'deleted' not in app:
                app['deleted'] = {'folders': 0, 'files': 0}

            deleted=app['deleted']                

            if app_status['aip'] >= Status.SOURCE_CLEAN_END:
                continue 

            if type(deleted['folders']) is str:
                deleted['folders']=0

            if type(deleted['files']) is str:
                deleted['files']=0

            deleted_files = deleted['files']
            deleted_folders = deleted['folders']

            app['status']['aip'] = app['status']['highlight'] = Status.SOURCE_CLEAN_START

            log_folder=abspath(f'{config.log_folder}/{config.project_name}/{app_name}/cleanup')
            create_folder(log_folder)

            cleanup_log_file= abspath(f"{log_folder}/{file_suffix}.log")
            cls.cleanup_log = Logger('File',level=config.log_level,file_name=cleanup_log_file,console_output=False)
            cls.cleanup_log.info(f'{config.project_name}/{app}')
            # cls._log.info(f'Cleanup file log: {cleanup_log_file}')

            app_folder = abspath(f'{config.stage_folder}\\{config.project_name}\\{app_name}')

            # cls._log.info(f'Reviewing {app} ({app_folder})')
            # with open (clean_up_log_file, 'a+') as file1: 
            #     with open (clean_up_log_folder, 'a+') as file2: 

            s=''                            

            cls.cleanup_log.info('Cleaning Folders')
            for subdir, dirs, files in walk(app_folder):
                for dir in dirs:
                    cls.cleanup_log.info(f'{subdir}\\{dir}')
                    if cls.find_with_list(dir,folder_list):
                        folder=join(subdir, dir)
                        rmtree(folder)
                        deleted_folders+=1
                        deleted['folders'] += 1
                        cls.cleanup_log.info(f'Removing Folder: {folder}')
                        found=True
                print(cls.show_progress())

            cls.cleanup_log.info('Cleaning Files')
            for subdir, dirs, files in walk(app_folder):
                for file in files:
                    cls.cleanup_log.info(f'{subdir}\\{dir}')
                    if cls.find_with_list(file,files_list):
                        file=join(subdir, file)
                        remove(file)
                        deleted_files+=1
                        deleted['files'] += 1
                        cls.cleanup_log.info(f'Removing File: {file}')
                        found=True
                print(cls.show_progress())

            # cls._log.info(f'Removed {deleted_files} files and {deleted_folders} folders from {app_folder}')
            # cls.log.info(f'Application {app}, removed {deleted_files} files and {deleted_folders} folders')               
            # cls.log.info(f'Cleanup log: {cleanup_log_file}\n')
    
            print(cls.show_progress())

        for app in apps:
            app_status = app['status']
            app_status['aip'] = app_status['highlight'] = Status.SOURCE_CLEAN_END

        config._save()
        cls.show_progress(done=True)
        config.log.info('Source Code cleanup done')
        return True

    def find_with_list(cls,find_in:str,pattern:list):
        rslt = False
        for p in pattern:
            try:
                if re.match(p,find_in):
                    rslt = True
                cls.cleanup_log.info(f'looking in {find_in} folder for {p} ({rslt})')
                if rslt:
                    break

            except re.error as ex:
                cls._log.warning(f'{ex.msg} for pattern {ex.pattern}')
            
        return rslt

    def get_title(cls) -> str:
        return "CLEANUP"

def rmtree(top):
    for root, dirs, files in walk(top, topdown=False):
        for name in files:
            filename = join(root, name)
            chmod(filename, S_IWUSR)
            remove(filename)
        for name in dirs:
            rmdir(join(root, name))

    chmod(top, S_IWUSR)
    rmdir(top)    

class cleanUpHL(Cleanup):
    def __init__(cls,config:Config, log_level:int):
        super().__init__(config,cls.__class__.__name__,log_level)

    @property
    def cleanup_file_prefix(cls):
        return "HL"
    
    def get_title(cls) -> str:
        return "CLEANUP FOR CAST HIGHLIGHT"



import pandas
from oneclick.discovery.sourceValidation import SourceValidation 
from pandas import DataFrame,read_excel
from os.path import abspath,exists
from oneclick.configTest import Config,Status
from cast_common.util import convert_LOC,list_to_text

from docx import Document
from docx.enum.text import WD_ALIGN_PARAGRAPH,WD_LINE_SPACING
from docx.shared import Length
from tqdm import tqdm

#todo: add totals to the cloc report (d1)
#todo: if SQL problems, add bullet under SQL Delivery to describe (d1)

class DiscoveryReport(SourceValidation):
    @property
    def name(self) -> str:
        return f'Discovery Report'
    def choose(self) -> bool:
        return True

    def __init__(cls):
        pass

    def cloc_report(cls,file:str,sheet:str):
        
        # read by 'Stats Before Code CleanUP' sheet of an Cloc_Output excel file
        df = DataFrame()
        try:
            if exists(file):
                df = read_excel(file,sheet_name=sheet)
        except ValueError as ex:
            cls.config.log.warning(f'No cloc information found for {sheet}')

        return df


    def run(cls):
        config = cls.config

        cloc_report = abspath(f'{config.report_folder}/{config.project_name}/{config.project_name}-cloc.xlsx')
        discovery_report = abspath(f'{config.report_folder}/{config.project_name}/{config.project_name}-source-code-discovery.docx')
        # create an instance of a word document
        doc = Document()
        # add a heading of level 0 (largest heading)
        doc.add_heading('Source code discovery report', 0)
        #doc_para = doc.add_paragraph(f'Please find below the completed discovery for the Project {config.project_name}')
        doc_para = doc.add_paragraph(f'Please find below the completed discovery for Project {config.project_name}. Note that we used CLOC for quick discovery and completeness verification. The actual lines of code discovered by the CAST solution may differ slightly from below. ')
        doc.add_heading('Questions', 1)
        doc.add_paragraph('<Specific questions to be added manually if needed. Remove this section if no questions>')

        doc.add_heading('Observation by Application', 1)

        print(cls.show_progress())

        for appl in config.applist:
            app_name = appl['name']
            # cls._log.info(f'Running {cls.__class__.__name__} for {app_name}')

            appl['status']['aip']=appl['status']['highlight']=Status.DISCOVERY_REPORT_START

            sql_report = abspath(f'{config.report}/{config.project_name}/{app_name}/{app_name}-SQLReport.xlsx')

            doc.add_heading(f'Application {app_name}:', 2)

            before_df = cls.cloc_report(cloc_report,f'{app_name}-Before')
            if before_df.empty:
                doc.add_paragraph(f'This application contains NO source code')
            else:
                # read by 'Stats Before Code CleanUP' sheet of an Cloc_Output excel file
                before_df=before_df[before_df['LANGUAGE'] != 'Totals']
                before_df[before_df.columns]=before_df.apply(lambda x: x.str.strip() if isinstance(x, str) else x).copy()
                #print(before_df)

                filt=(before_df['APPLICABLE']==False)
                l=before_df.loc[filt,['LANGUAGE','CODE']].sort_values(by=['CODE'], ascending=False) 
                non_code=''
                if len(l) == 1:
                    str(l.iloc[0]['CODE']//1000) +' KLOC of '+ l.iloc[0]['LANGUAGE']
                elif len(l) > 1:
                    non_code= str(l.iloc[0]['CODE']//1000) +' KLOC of '+ l.iloc[0]['LANGUAGE'].strip() +' non code files and ~'+ str(l.iloc[1]['CODE']//1000) +' KLOC of '+ l.iloc[1]['LANGUAGE']

                # read by 'Stats After Code CleanUP' sheet of an Cloc_Output excel file
                after_df = cls.cloc_report(cloc_report,f'{app_name}-After')
                #print(after_df)

                # read by 'Summary' sheet of an SQL_Output excel file
                if exists(sql_report):
                    sql_df = read_excel(sql_report,sheet_name='Summary')
                else:
                    sql_df = DataFrame(columns=['Name','Total','Unique','Dups'])
                    # cls._log.warning('No SQL found')
                #sql_df = sql_df[sql_df['Total']>0]

                # out of scope code base
                langs=len(before_df)
                total, unit = convert_LOC(int(before_df['CODE'].sum()))
                doc.add_paragraph(f"This delivery contains {langs} discovered technolog{'ies' if langs > 1 else 'y'}/extension{'s' if langs > 1 else ''}  and a total of {total} {unit}.",style='List Bullet')

                bsuport = after_df[after_df['APPLICABLE']==True]
                total = int(bsuport['CODE'].sum())
                total, unit = convert_LOC(total)

                lang_list = list(bsuport['LANGUAGE'].str.strip())
                langs = len(lang_list)
                if langs >1:
                    lang_list[-1]=f'and {lang_list[-1]}'
                bsup_lang = ', '.join(lang_list)
                doc.add_paragraph(f"{len(bsuport)} technolog{'ies' if langs > 1 else 'y'}/extension{'s are' if langs > 1 else ' is'} relevant for the CAST analysis, which include {bsup_lang}.",style='List Bullet')

                nsuport = after_df[after_df['APPLICABLE']==False]
                total = int(bsuport['CODE'].sum())
                total, unit = convert_LOC(total)

                lang_list = list(nsuport['LANGUAGE'].str.strip())
                langs = len(lang_list)
                if langs > 1:
                    lang_list[-1]=f'and {lang_list[-1]}'
                bsup_lang = ', '.join(lang_list)
                doc.add_paragraph(f"The remaining technolog{'ies' if langs > 1 else 'y'}/extension{'s are' if langs > 1 else ' is'} not relevant and will not be analyzed. These include {bsup_lang}.",style='List Bullet')

                # in scope code base
                # asuport = after_df[before_df['APPLICABLE']==True]
                # total = int(asuport['CODE'].sum())
                # total, unit = convert_LOC(total)
                #doc.add_paragraph(f'After removing all Sample, Test and other non production related code there is ({total} {unit}) in scope for this project',style='List Bullet 2')
                #print(sql_df)
                create_df=sql_df[sql_df['Name'].str.startswith('Create')]
                if len(create_df):
                    sql_items = []
                    create_df.sort_values(['Unique'],ascending=False,inplace=True)
                    for index,item in create_df.iterrows():
                        count = int(item['Unique'])             # get the unique count from dataframe
                        name = item['Name'].split(" ")[1]       # remove the word Create
                        if count == 1: name = name[:-1]         # only one item, remove make name singular
                        sql_items.append(f'{count} {name}')
                    doc.add_paragraph(f'The database contains a total of {list_to_text(sql_items)}.',style='List Bullet')
                else:
                    doc.add_paragraph(f'No database objects found.',style='List Bullet')

                removed_code=before_df['CODE'].sum()-after_df['CODE'].sum()
                doc.add_paragraph(f'We removed {round(removed_code/1000,1)} KLOC of sample, test and other non-production code from the source code.',style='List Bullet')
                after_df=after_df.drop(after_df[after_df['APPLICABLE']==0.0].index)
                total = after_df['CODE'].sum()//1000
                doc.add_paragraph(f"Here is a high-level summary of the source code that will be analyzed by CAST, totaling {total} KLOC.",style='List Bullet')

                after_df = after_df[['LANGUAGE','FILES','CODE']]
                after_df = after_df[:-1]
                files_count=after_df['FILES'].sum()
                code_count=after_df['CODE'].sum()
                new_row = pandas.DataFrame({'LANGUAGE':'Totals', 'FILES':[files_count], 'CODE':[code_count]})
                after_df = pandas.concat([after_df,new_row])
                
                t = doc.add_table(after_df.shape[0]+1, after_df.shape[1])
                # add the header rows.
                for j in range(after_df.shape[-1]):
                    t.cell(0,j).text = after_df.columns[j].strip()
                    if j > 0:
                        t.cell(0,j).paragraphs[0].alignment=WD_ALIGN_PARAGRAPH.RIGHT

                # add the rest of the data frame
                for i in range(after_df.shape[0]):
                    for j in range(after_df.shape[-1]):
                        number = after_df.values[i,j]
                        if type(number) is int:
                            number = "{:,}".format(number)
                        t.cell(i+1,j).text = number
                        if j > 0:
                            t.cell(i+1,j).paragraphs[0].alignment=WD_ALIGN_PARAGRAPH.RIGHT
                make_rows_bold(t.rows[len(after_df)])

                # Adding style to a table
                t.style = 'Medium Shading 1 Accent 1'

            appl['status']['aip']=appl['status']['highlight']=Status.DISCOVERY_REPORT_END

        # now save the document to a location
        doc.save(discovery_report)

        print(cls.show_progress(clear=True))
        config.log.info(f'Source Code Discovery report generated at {discovery_report} (note: pls review report document before sharing)\n')
        print(cls.show_progress(done=True))
        return True

    def get_title(cls) -> str:
        return "DELIVERABLE DISCOVERY REPORT"

def make_rows_bold(*rows):
    for row in rows:
        for cell in row.cells:
            for paragraph in cell.paragraphs:
                for run in paragraph.runs:
                    run.font.bold = True



from oneclick.configTest import Config,Status
from cast_common.logger import Logger,INFO
from cast_common.util import run_process,check_process,format_table,create_folder
from oneclick.discovery.sourceValidation import SourceValidation 
from time import sleep

from platform import system
from os import remove
from os.path import exists,abspath,getsize
from re import findall
from pandas import DataFrame,ExcelWriter
from openpyxl import load_workbook
from tqdm import tqdm
from subprocess import Popen

from string import ascii_uppercase
from win32api import GetLogicalDriveStrings
from win32wnet import WNetOpenEnum,WNetEnumResource
from win32netcon import RESOURCE_REMEMBERED,RESOURCETYPE_DISK

from ctypes import windll, c_int, c_wchar_p
from time import sleep
from openpyxl import load_workbook

#TODO: Convert total line to formulas (d1-SHP)
#TODO: Format all numbers as integers not text (d1-SHP)
#TODO: Group tabs in pairs (before, after) then by application (d2)

class ClocPreCleanup(SourceValidation):
    writer = None

    @property
    def max_cloc(cls):
        return ClocPreCleanup._max_cloc
    @property
    def running_cloc(cls):
        return ClocPreCleanup._running_cloc
    @running_cloc.setter
    def running_cloc(cls,value):
        ClocPreCleanup._running_cloc = value

    @property
    def choose(self) -> bool:
        return True
    
    @property
    def name(self) -> str:
        return 'Technology Discovery - Before Cleanup'

    def __init__(cls):
        ClocPreCleanup._max_cloc=2
        ClocPreCleanup._running_cloc=0
        cls._df = {}
        pass

    @property
    def phase(cls):
        return 'Before'

    @property
    def cloc_base(cls):
        return f'{cls.config.base}\\cloc' 
    @property
    def cloc_project(cls):
        return f'{cls.cloc_base}\\{cls.config.project_name}'

    @property
    def cloc_results(cls):
        return cls._df

    def cloc_output_path(cls,appl:str):
        return abspath(f'{cls.config.report_folder}/{cls.config.project_name}/{appl}/{appl}-cloc-{cls.phase}.txt')

    def cloc_output_ignore_path(cls,appl:str):
        return abspath(f'{cls.config.report_folder}/{cls.config.project_name}/{appl}/{appl}-cloc-ignored-{cls.phase}.txt')

    def _get_free_drive(cls):
        drives = set(ascii_uppercase[2:])
        for d in GetLogicalDriveStrings().split(':\\\x00'):
            drives.discard(d)
        # Discard persistent network drives, even if not connected.
        henum = WNetOpenEnum(RESOURCE_REMEMBERED, 
            RESOURCETYPE_DISK, 0, None)
        while True:
            result = WNetEnumResource(henum)
            if not result:
                break
            for r in result:
                if len(r.lpLocalName) == 2 and r.lpLocalName[1] == ':':
                    drives.discard(r.lpLocalName[0])
        if drives:
            return sorted(drives)[-1] + ':'

    def _run_cloc(cls,work_folder:str,cloc_output:str,cloc_output_ignored:str):
        args = [cls.config.cloc,work_folder,"--report-file",cloc_output,"--ignored",cloc_output_ignored,"--quiet"]
        cls.log.debug(' '.join(args))
        proc = run_process(args,False)
        if proc.poll() is not None and exists(cloc_output):
            return 'DONE'
        else:
            return proc

    def run(cls):

        config = cls.config
        # cls.open_excel_writer()
        
        #cls.cloc_path=abspath(f'{config.base}\\scripts\\{cls.config.cloc_version}')
        
        list_of_tech_file=abspath(f'{config.scripts_folder}/ListOfTechnologies.csv')
        with open(list_of_tech_file) as f:
            cls.tech_list = f.read().splitlines()
            
        process = {}
        output = {}
        cloc_run=False

        # create a subst drive
        DefineDosDevice = windll.kernel32.DefineDosDeviceW
        DefineDosDevice.argtypes = [ c_int, c_wchar_p, c_wchar_p ]

        project_folder=abspath(f'{config.stage_folder}/{config.project_name}')
        # Create a subst. Check the return for non-zero to mean success
        drive=project_folder
        platform = system()
        if platform == 'Windows':
            drive = cls._get_free_drive()
            if drive is None or DefineDosDevice(0, drive, project_folder ) == 0:
                raise RuntimeError("Subst failed")

        for appl in config.applist:
            app_status = appl['status']['aip']

            # was there an error with the last run
            # if so we need to reset the status to restart cloc
            if cls.phase=='Before' and app_status == Status.CLOC_PRE_CLEAN_ERROR:
                appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_START
            elif cls.phase=='After' and app_status == Status.CLOC_POST_CLEAN_ERROR:
                appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_POST_CLEAN_START

        print(cls.show_progress())

        create_folder(abspath(f'{config.report_folder}/{config.project_name}'))

        # if cls.phase=='Before':
        #     ClocPreCleanup.writer = ExcelWriter(ClocPreCleanup.excel_file_name, engine='xlsxwriter')

        # do a little cleanup        
        for appl in config.applist:
            app_name = appl['name']
            app_status = appl['status']['aip']
            cloc_output = cls.cloc_output_path(app_name)
            cloc_output_ignored = cls.cloc_output_ignore_path(app_name)

            # if config.start== cls.name:
            #     if cls.phase=='Before':
            #         appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_REPORT
            #     else:
            #         appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_POST_CLEAN_START

            if (cls.phase=='Before' and (app_status == Status.CLOC_PRE_CLEAN_QUEUE or app_status < Status.CLOC_PRE_CLEAN_START)) or \
                (cls.phase=='After' and (app_status == Status.CLOC_POST_CLEAN_QUEUE or app_status < Status.CLOC_POST_CLEAN_START)) :
                if exists(cloc_output):
                    cls.config.log.info(f'Removing {cloc_output}')
                    remove(cloc_output)
                    print(cls.show_progress())
                if exists(cloc_output_ignored):
                    cls.config.log.info(f'Removing {cloc_output_ignored}')
                    remove(cloc_output_ignored)
                    print(cls.show_progress())
                if 'loc' in appl:
                    appl['loc']=''

        print(cls.show_progress())

        #run cloc for all applications
        while True:
            for appl in config.applist:
                app_name = appl['name']
                app_status = appl['status']['aip']

                if (cls.phase == 'Before' and app_status >= Status.CLOC_PRE_CLEAN_END) or \
                    (cls.phase == 'After' and app_status >= Status.CLOC_POST_CLEAN_END):
                        continue

                create_folder(f'{config.report_folder}/{config.project_name}/{app_name}')
                cloc_output = cls.cloc_output_path(app_name)
                cloc_output_ignored = cls.cloc_output_ignore_path(app_name)
                if platform == 'Windows':            
                    work_folder = abspath(f'{drive}/{app_name}')
                else:
                    work_folder = abspath(f'{config.work}/{config.project_name}/{app_name}')

                #if the report is already out there - no need to continue
                if (cls.phase=='Before' and app_status in [Status.CLOC_PRE_CLEAN_QUEUE,Status.CLOC_PRE_CLEAN_START,Status.UNPACK_END]) or \
                   (cls.phase=='After' and app_status in [Status.CLOC_POST_CLEAN_QUEUE,Status.CLOC_POST_CLEAN_START,Status.SOURCE_CLEAN_END]) :
                    #this application cloc run has been queued, can we run it now?
                    if cls.running_cloc < cls.max_cloc:
                        # if it exists, remove the report file

                        proc = cls._run_cloc(work_folder, cloc_output, cloc_output_ignored)
                        if proc == 'DONE':
                            appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_REPORT if cls.phase=='Before' else Status.CLOC_POST_CLEAN_REPORT
                            config._save()
                        else:
                            cls.running_cloc += 1
                            process[app_name] = proc
                            appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_RUNNING if cls.phase=='Before' else Status.CLOC_POST_CLEAN_RUNNING
                            config._save()

                elif (cls.phase=='Before' and app_status == Status.CLOC_PRE_CLEAN_RUNNING) or \
                     (cls.phase=='After' and app_status == Status.CLOC_POST_CLEAN_RUNNING):
                    #cloc is running fvor this application, check if it has completed yet
                    try:
                        #get the results
                        if (app_name in process and process[app_name].poll() is None):
                            pass
                        else:
                            if exists(cloc_output) and getsize(cloc_output) > 0:
                                if cls.phase == 'Before':
                                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_REPORT
                                elif cls.phase == 'After':
                                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_POST_CLEAN_REPORT
                                # process[app_name]='DONE'
                                cls.running_cloc -= 1
                            else:
                                if cls.phase == 'Before':
                                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_ERROR
                                elif cls.phase == 'After':
                                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_POST_CLEAN_ERROR

                                output, error = process[app_name].communicate()

                                print(cls.show_progress(done=True))
                                config.log.error(output.decode())

                            config._save()

                    except IOError:
                        if not exists(cloc_output) and getsize(cloc_output) == 0:
                            cls.log.error(f'Error running cloc on {cloc_output}')
                    except ValueError as ex:
                        if str(ex) == 'read of closed file':
                            cls.log.debug(f'Process already closed {app_name}')
                    except Exception as ex:
                        cls.log.warning(f'{type(ex)} Error: {str(ex)}')
                        pass

                if cls.phase == 'Before' and app_status == Status.CLOC_PRE_CLEAN_REPORT and not exists(cloc_output):
                    #something went wrong with the cloc run
                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_ERROR
                    continue
                if cls.phase == 'After' and app_status == Status.CLOC_POST_CLEAN_REPORT and not exists(cloc_output):
                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_POST_CLEAN_ERROR
                    continue

                if (cls.phase == 'Before' and app_status == Status.CLOC_PRE_CLEAN_REPORT) or \
                   (cls.phase == 'After' and app_status == Status.CLOC_POST_CLEAN_REPORT) :
                    cls.process_results()
                    appl['status']['aip'] = appl['status']['highlight'] = Status.CLOC_PRE_CLEAN_END if cls.phase=='Before' else Status.CLOC_POST_CLEAN_END
                print(cls.show_progress())
                sleep(1)
            print(cls.show_progress())

            # are we done with all applications?
            done = True
            for appl in config.applist:
                app_status = appl['status']['aip']
                if (cls.phase == 'Before' and app_status < Status.CLOC_PRE_CLEAN_END) or \
                   (cls.phase == 'After' and app_status < Status.CLOC_POST_CLEAN_END) :
                    done = False
                    break
            if done:
                break

        # Delete the subst.
        if platform == 'Windows' and DefineDosDevice(2, drive, project_folder ) == 0:
            raise RuntimeError("Subst failed")

        if cls.phase == 'After' and ClocPreCleanup.writer is not None:
            ClocPreCleanup.writer.close()


        config._save()
        cls.log.info(cls.show_progress(True))

        cloc_location = abspath(f'{config.report_folder}/{config.project_name}')
        cls.log.info(f'All pre cleanup CLOC reports generated for project: {config.project_name}, see {cloc_location}')
        return True

    def process_results(self):

        def parse_rpt(app_name,cloc_output,cloc_output_ignored,phase):
            with open(cloc_output, 'r') as f:
                content = f.read()

            #extracting required data from content of cloc_output.txt using python regex
            header=content.split('\n')[2]
            header_list=findall('\w+',header.upper())

            summary='\n'.join(content.split('\n')[4:-4])
            pattern='(.{25})\s{1,}(\d{1,})\s{1,}(\d{1,})\s{1,}(\d{1,})\s{1,}(\d{1,})'
            statistics_list=findall(pattern,summary)
            with open(cloc_output_ignored, 'r') as fp:
                lines = len(fp.readlines())
            statistics_list.append(('Unknown Files','0','0','0',str(lines)))
            df = DataFrame(statistics_list,columns=header_list)

            #making technolgy case insensitive
            self.tech_list = list(map(lambda x: x.lower().strip(), self.tech_list))
            df['APPLICABLE']=df['LANGUAGE'].str.lower().str.strip().isin(self.tech_list)

            #converting column values into int from string
            numbers=['FILES','BLANK','COMMENT','CODE']
            total_line=['']
            for name in numbers:
                df[name] = df[name].astype('int')
            tab_name = f'{app_name}-{phase}'
            tab_name = (tab_name[:30] + '..') if len(tab_name) > 30 else tab_name
            loc = df['CODE'].sum()
            appl['loc']= f'{loc:>10,.0f}'
            return df,tab_name


        config = self.config
        excel_file_name = abspath(f'{config.report_folder}/{config.project_name}/{config.project_name}-cloc.xlsx')
        if exists(excel_file_name):
            remove(excel_file_name)
        writer = ExcelWriter(excel_file_name, engine='xlsxwriter')

        for appl in config.applist:
            status = appl['status']['aip']
            app_name = appl['name']

            if status >= Status.CLOC_PRE_CLEAN_REPORT:
                before_cloc_rpt = abspath(f'{config.report_folder}/{config.project_name}/{app_name}/{app_name}-cloc-Before.txt')
                before_ignr_rpt = abspath(f'{config.report_folder}/{config.project_name}/{app_name}/{app_name}-cloc-ignored-Before.txt')
                after_cloc_rpt = abspath(f'{config.report_folder}/{config.project_name}/{app_name}/{app_name}-cloc-After.txt')
                after_ignr_rpt = abspath(f'{config.report_folder}/{config.project_name}/{app_name}/{app_name}-cloc-ignored-After.txt')

                if status >= Status.CLOC_PRE_CLEAN_REPORT and exists(before_cloc_rpt) and exists(before_ignr_rpt):
                    df,tab_name = parse_rpt(app_name, before_cloc_rpt, before_ignr_rpt, 'Before')
                    format_table(writer, df, tab_name, total_line=True)

                if status >= Status.CLOC_POST_CLEAN_REPORT and exists(after_cloc_rpt) and exists(after_ignr_rpt):
                    df,tab_name = parse_rpt(app_name, after_cloc_rpt, after_ignr_rpt, 'After')
                    format_table(writer, df, tab_name, total_line=True)

        writer.close()
        pass

class ClocPostCleanup(ClocPreCleanup):
    @property
    def choose(self) -> bool:
        return False
    
    @property
    def name(self) -> str:
        return 'Technology Discovery - After Cleanup'

    def __init__(cls):
        pass
    
    @property
    def phase(cls):
        return 'After'

        


__author__ = "Nevin Kaplan"
__email__ = "n.kaplan@castsoftware.com"
__copyright__ = "Copyright 2024, CAST Software"

from pkg_resources import get_distribution
from oneclick.configTest import Config,Status
from oneclick.base import Base
from oneclick.exceptions import InvalidConfiguration,InvalidConfigNoBase

from oneclick.discovery.unzip import Unzip
from oneclick.discovery.prep import Prepare
from oneclick.discovery.cloc import ClocPreCleanup,ClocPostCleanup
from oneclick.discovery.profiler import ProfilerPreCleanup
from oneclick.discovery.cleanup import Cleanup,rmtree
from oneclick.discovery.sqlDiscovery import SQLDiscovery
from oneclick.discovery.discoveryReport import DiscoveryReport



from oneclick.analysis.analysis import Analysis
from oneclick.analysis.runAnalysis import RunHighlight
from oneclick.analysis.highlight_analysis import HLAnalysis
from oneclick.analysis.aip_analysis import AIPAnalysis
from oneclick.analysis.trackAnalysis import TrackAnalysis

from inquirer import List,prompt
from os.path import abspath,exists

if __name__ == '__main__':
    version = get_distribution('com.castsoftware.uc.oneclick').version
    print(f'\nCAST OneClick Setup, v{version}')
    print(f'com.castsoftware.uc.python.common v{get_distribution("com.castsoftware.uc.python.common").version}')
    print('Copyright (c) 2024 CAST Software Inc.')
    print('If you need assistance, please contact oneclick@castsoftware.com')
    
    config = Config()
    Base(config)    
    try:
        config.log.info(f'****************** Starting CAST OneClick for project {config.project_name} ******************')

        if config.reset:
            config.log.info(f'Resetting all applications')
            for app in config.applist:
                app['status']['highlight'] = app['status']['aip'] = Status.NOT_STARTED
            config._save()

            for folder in [abspath(f'{config.stage_folder}/{config.project_name}'),
                        abspath(f'{config.highlight_folder}/{config.project_name}'),
                        abspath(f'{config.report_folder}/{config.project_name}')]:
                config.log.info(f'Removing {folder}')
                if exists(folder):
                    rmtree(folder)
            for app in config.applist:
                app['deleted'] = {'folders':'','files':''}
                app['sql'] = {'tables':'','functions':'','procedures':'','views':'','triggers':''}
                app['loc'] = ''
                app['unpacked'] = ''

        # config.log.info(config.application_report())
        process = [

            # environment setup
            Prepare(),
            Unzip(),
            
            # source code preparation
            ClocPreCleanup(),
            Cleanup(),
            ClocPostCleanup(),
            # ProfilerPreCleanup(),

            # reporting
            SQLDiscovery(),
            DiscoveryReport(),

            #todo: add discovery report email notification
            # EmailNotification(),

            # application analysis process
            #TODO DLM module?
            # AIPAnalysis(),
            # HLAnalysis(),
            # TrackAnalysis()

            RunHighlight()

            #RunARG(config,log_level)

        ]

        choices = []
        for obj in process:
            if  obj.choose:
                choices.append(f'{obj.name}') 

        start=0
        end=len(choices)-1
        if not (config.quiet):
            questions = [
                List(
                    'start with step',
                    message="Starting process step?", 
                    choices=choices,
                    default=choices[0]
                    ),
                List(
                    'End with step',
                    message="Ending process step", 
                    choices=choices,
                    default=choices[len(choices)-1]
                    )
            ]
            answers = prompt(questions)
            config.start = answers['start with step']
            config.end = answers['End with step']

        start = end = 0        
        if config.start == None or len(config.start) == 0:
            start=0
        else:
            start = 0
            for i in range(len(process)):
                if process[i].name.startswith(config.start):
                    start=i
                    break

        if config.end == None or len(config.end) == 0:
            end=len(process)-1
        else:
            end = len(process)
            for i in range(len(process)):
                if process[i].name.startswith(config.end):
                    end=i
                    break

        if start > end:
            raise InvalidConfiguration(f'End step {process[end].name} is before start step {process[start].name}')
        elif not process[start].choose:
            raise InvalidConfiguration(f'Start step {process[start].name} is not a selectable step')

        config.start = process[start].name
        config.end = process[end].name

        config.log.info(f'Starting project {config.project_name} at "{config.start}" and ending at "{config.end}"')
        active=False
        step = 0
        for p in process:
            step += 1
            #look for starting class
            if p.name == process[start].name:
                active=True

            if not active: 
                config.log.info(f'Skipping {p.name}')
                continue

            # if issubclass(type(p), Analysis) and end == 'Discovery':
            #     config.log.info('--end Discovery selected, process stopping here')
            #     break

            p.log.info(f'=> STEP {step} : {p.name}')
            
            if p.can_run:
                # config.log.info(f'Running {p.name}')
                status = p.run()
                if not status:
                    # config.log.error(f'Error running {p.name}')
                    break
                config._save()
            else:
                config.log.info(f'Skipping {p.name}, configuration is incomplete')
            pass


    except InvalidConfiguration as ic:
        config.log.error(ic)
    except Exception as ex:
        from traceback import format_exc

        print ('\n\n\n\n\n\n\n\n')

        config.log.error(format_exc())
    config.log.info(f'Thanks for using CAST OneClick. For more information, contact oneclick@castsoftware.com')




from oneclick.discovery.sourceValidation import SourceValidation 
from oneclick.configTest import Config,Status
from os import walk
from os.path import abspath
from re import findall,IGNORECASE,sub

from cast_common.util import  format_table
from pandas import DataFrame,ExcelWriter,Series,json_normalize


#TODO: Add filename and location for each item (d1)
class SQLDiscovery(SourceValidation):
    @property
    def choose(self) -> bool:
        return True
    
    @property
    def name(self) -> str:
        return 'SQL Discovery'

    def __init__(cls):
        pass

    _pattern={
        'Create Tables'      :r'create\stable\s[^#]([^\n|^\s|^\(]+)[^\(]',
        'Create Functions'   :r'create\sfunction\s([^\n|^\s|^\(]+)[^\(]',
        'Create Procedures'  :r'create\sprocedure\s([^\n|^\s|^\(]+)[^\(]',
        'Create Views'       :r'create\sview\s([^\n|^\s|^\(]+)[^\(]',
        'Create Triggers'    :r'create\strigger\s([^\n|^\s|^\(]+)[^\(]',
        'Alter Tables'       :r'alter\stable\s([^\n|^\s|^\(]+)',
        'Alter Functions'    :r'alter\sfunction\s([^\n|^\s|^\(]+)',
        'Alter Procedures'   :r'alter\sprocedure\s([^\n|^\s|^\(]+)',
        'Alter Views'        :r'alter\sview\s([^\n|^\s|^\(]+)',
        'Alter Triggers'     :r'alter\strigger\s([^\n|^\s|^\(]+)',
    }

    def parse_sql(cls,file):

        rslt = {}
        rslt['file-name']=file
        with open(file, encoding="cp437") as f:
            content = f.read()
            for key in cls._pattern.keys():
                pattern = cls._pattern[key]
                rslt[key]=findall(pattern,content,IGNORECASE)
                rslt[key]=list(map(lambda x: sub(r'\W+','',x), rslt[key]))
                if len(rslt[key])==0:
                   rslt[key]=[] 
        cls._data.append(rslt)
        pass

    def run(cls):
        config = cls.config
        apps= config.applist

        print(cls.show_progress())

        for app in apps:
            app_name=app['name']
            cls.log.info(f'application {app_name}')

            status = app['status']
            if status['aip'] < Status.CLOC_POST_CLEAN_END:
                cls.log.warning(f'CLOC not completed for {app_name}')
                return False

            status['aip'] = status['highlight'] = Status.SQL_DISCOVERY_START
            if 'sql' not in app:
                app['sql']={'tables':0,'functions':0,'procedures':0,'views':0,'triggers':0}
            config._save()

            cls._data=[]
            sql_files = []
            non_sql_files = []
            app_folder = abspath(f'{config.stage_folder}/{config.project_name}/{app_name}')
            for root, dirs, files in walk(app_folder):
                for file in files:
                    if file.endswith(".bod") or \
                    file.endswith(".fnc") or \
                    file.endswith(".prc") or \
                    file.endswith(".trg") or \
                    file.endswith(".bdy") or \
                    file.endswith(".spc") :
                        fn = abspath(f'{root}/{file}')
                        cls.log.warning(f'Non-standard SQL file found: {fn}')
                        non_sql_files.append(fn)
                        # cls._log.warning(f'Potential SQL files with another alternate extension found in {app} review SQLReport and rename if appropriate')

                    if file.endswith(".sql") or file.endswith(".dtd"):
                        sql_files.append(abspath(f'{root}/{file}'))
                pass

            # cls._log.info(f'Found {len(sql_files)} SQL and {len(non_sql_files)} potential SQL files.')

            if len(sql_files):
                for file in sql_files:
                    cls.parse_sql(file)

                summary_df = DataFrame(columns=['Name','Total','Unique','Dups'])
                df = json_normalize(cls._data)
                detail = {}
                for key in cls._pattern.keys():
                    if key in df.keys():
                        detail_df=df.explode(key).dropna()
                        if not detail_df.empty:
                            detail_df=detail_df[['file-name',key]]
                            detail[key]=detail_df.sort_values(by=[key])
                            dups = len(detail_df.drop_duplicates(subset=[key]))
                            total=len(detail_df)
                            summary_df.loc[len(summary_df.index)] = [key,total,dups,total-dups]
                        else:
                            summary_df.loc[len(summary_df.index)] = [key,0,0,0]
                            detail[key]=None

                app['sql']['tables'] = int(summary_df.loc[0]['Total'])
                app['sql']['functions'] = int(summary_df.loc[1]['Total'])
                app['sql']['procedures'] = int(summary_df.loc[2]['Total'])
                app['sql']['views'] = int(summary_df.loc[3]['Total'])
                app['sql']['triggers'] = int(summary_df.loc[4]['Total'])
                config._save()
                
                # total_artifacts = total_tables + total_functions + total_procedures + total_triggers + total_views
                # print('-------------------------------------')
                # print('Artifacts                     Count')
                # print('-------------------------------------')
                # print(f'Tables                          {total_tables}')
                # print(f'Functions                       {total_functions}')
                # print(f'Procedures                      {total_procedures}')
                # print(f'Views                           {total_views}')
                # print(f'Triggers                        {total_triggers}')
                # print('-------------------------------------')
                # print(f'Total Artifacts                 {total_artifacts}')
                # print('-------------------------------------')

                if not summary_df.empty:
                    summary_df=summary_df.sort_values(['Name'],ascending=False)
                    filename = abspath(f'{config.report_folder}/{config.project_name}/{app_name}/{app_name}-SQLReport.xlsx')
                    writer = ExcelWriter(filename, engine='xlsxwriter')
                    dups_format = writer.book.add_format({'bg_color': '#FFFF00', 'font_color': '#9C0006'})
                    xls=format_table(writer,summary_df,'Summary',total_line=True)
                    xls.conditional_format(f'D2:D{len(summary_df)}', {'type':'cell','criteria':'>','value':0,'format':dups_format})

                    for key in cls._pattern.keys():
                        if not detail[key] is None:
                            xls = format_table(writer,detail[key],key)
                            xls.conditional_format(f'A1:B{len(detail[key])}', {'type':'formula','criteria':'=COUNTIF($B:$B,$B1)>1','format':dups_format})

                    writer.close()
                    print(cls.show_progress())
                    # cls._log.info(f'SQL Discovery Report: {filename}')
                    # cls._log.info(f'detailed discovery report available at {filename}\n')
            else:
                #cls._log.warning(f'No SQL found\n')
                pass
            pass
        # cls._log.info('SQLDiscovery complete.')
        pass
        print(cls.show_progress())
        return True





from cast_common.logger import Logger,INFO,DEBUG
from importlib import import_module
from oneclick.configTest import Config
from argparse import ArgumentParser
from argparse_formatter import FlexiFormatter,ParagraphFormatter

json = [
    {
        "module": "oneclick.discovery.prep",
        "class": "Prepare",
        "catagory": "SourceValidation",
        "log": INFO
    }

]


def get_argparse_defaults(parser):
    defaults = {}
    for action in parser._actions:
        if not action.required and action.dest != "help":
            defaults[action.dest] = action.default
    return defaults

parser = ArgumentParser(prog='OneClick',  formatter_class=lambda prog: FlexiFormatter(prog, width=99999, max_help_position=60))

config=Config(parser,get_argparse_defaults(parser))

for item in json:
    print(item)
    try: 
        module = import_module(item['module'])
        class_name = getattr(module, item['class'])
        catagory = item['catagory']
        log_level = item['log']

        instance = class_name(config,log_level)

    except (ImportError, AttributeError) as ex:
        print(ex)
        pass    





class InvalidConfiguration(RuntimeError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class InvalidConfigNoBase(InvalidConfiguration):
    pass

class NoConfigFound(RuntimeError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


from argparse import ArgumentParser

from os.path import abspath,exists,dirname
from os import chdir,getcwd,getenv
from cast_common.util import yes_no_input,folder_input,secret_input,url_input,string_input
from oneclick.config import Config
from oneclick.exceptions import NoConfigFound,InvalidConfiguration,InvalidConfigNoBase
from pkg_resources import get_distribution


from argparse import ArgumentParser
from pkg_resources import get_distribution
from oneclick.config import Config
from oneclick.exceptions import InvalidConfigNoBase

def main():
    version = get_distribution('com.castsoftware.uc.oneclick').version
    print(f'\nCAST OneClick Setup, v{version}')
    print('Copyright (c) 2024 CAST Software Inc.')
    print('If you need assistance, please contact oneclick@castsoftware.com')

    parser = ArgumentParser(prog='OneClick')
    parser.add_argument('-b', '--baseFolder', help='Base Folder Location', metavar='BASE_FOLDER')
    parser.add_argument('-w', '--workingFolder', help='Working Folder Location', metavar='WORKING_FOLDER')
    parser.add_argument('-d', '--debug', type=bool)

    try:
        config = Config(parser)
    except InvalidConfigNoBase as e:
        print(e)
        return

    configure_oneclick(config)

def configure_oneclick(config):
    chdir(config.base)

    print(f'OneClick installation location: {config.base}')
    config.workbase = input('OneClick Working Folder: ') or config.workbase

    config.java_home = configure_java(config)
    config.profiler = configure_profiler(config)

    if input('Configure MRI automatic scans? (y/n) ').lower() == 'y':
        configure_mri(config)

    if input('Configure Highlight automatic scans? (y/n) ').lower() == 'y':
        configure_highlight(config)

def configure_java(config):
    config.cloc_version='cloc-1.96.exe'

def configure_java(config):
    if len(config.java_home) == 0:
        config.java_home=getenv('JAVA_HOME')
    config.java_home = folder_input("Java installation folder location",config.java_home,'bin/java.exe')

def configure_profiler(config):
    profiler_folder = ''
    if len(config.profiler) > 0:
        profiler_folder = '\\'.join(config.profiler.split('\\')[:-1])
    config.profiler = folder_input("CAST-Profiler folder location",profiler_folder,'CAST-Profiler.exe',combine=True) 

def configure_mri(config):
    print ('Configuring for AIP Console')
    if yes_no_input('\tAre both "AIP Console Enterprise Edition" and the matching "AIP Console automation tools" installed?'):
        config.console_url = url_input('\t"AIP Console Enterprise Edition" URL',config.console_url)
        config.console_cli = folder_input('\t"AIP Console automation tools" location',dirname(config.console_cli),"aip-console-tools-cli.jar",True)
        config.console_key = secret_input('\t"AIP Console Enterprise Edition" API Key', config.console_key)
        config.enable_security_assessment = yes_no_input('\tEnable security analysis',config.enable_security_assessment)
        config.blueprint = yes_no_input('\tEnable Blueprint analysis',config.blueprint)

    if config.is_console_active:
        print ('AIP Console configuration complete')
    else:
        print ('Error updating AIP Console configuration!')
    print ('\n')

def configure_highlight(config):
    config.hl_cli = folder_input('\tHighlight command line interface installation location',config.hl_cli,'HighlightAutomation.jar',True)
    agent_folder = f'{config.perl_install_dir}'.replace('\\strawberry\\perl','')
    agent_folder = folder_input('\tHighlight agent installation location',agent_folder,'strawberry',False)
    config.perl_install_dir=abspath(f'{agent_folder}/strawberry/perl')
    config.analyzer_dir=abspath(f'{agent_folder}/perl')

    if yes_no_input('\tAre all of your applications using the same Highlight portal?'):
        if len(config.hl_url)==0:
            config.hl_url = "https://rpa.casthighlight.com/"
        config.hl_url = url_input('\t\tHighlight URL',config.hl_url)
        config.hl_url = f"{config.hl_url}"
        config.hl_user = string_input('\t\tHighlight user name',config.hl_user)
        config.hl_password = secret_input('\t\tHighlight password', config.hl_password)
        config.hl_instance = string_input('\t\tHighlight instance id',config.hl_instance)

    if config.is_hl_active:
        print ('Highlight configuration complete')
    else:
        print ('Error updating Highlight configuration!')
    print ('\n')

if __name__ == '__main__':
    main()




from oneclick.base import Base
from cast_common.logger import Logger
from json import dump
from oneclick.config import Config
from os.path import abspath

class SourceValidation(Base):
    _log=None

    def __init__(cls):
        # if SourceValidation._log is None:
        #     SourceValidation._log = Logger(name=log_name,level=log_level,file_name=config.log_filename)
        #     SourceValidation._log_level=log_level
        pass

    @property
    def required(cls):
        return True
    
    def get_title(cls) -> str:
        return cls.__class__.__name__




from oneclick.analysis.analysis import Analysis
from subprocess import Popen,PIPE
from oneclick.configTest import Config,Status
from cast_common.hlRestCall import HLRestCall
from cast_common.util import run_process,create_folder,check_process
from os.path import abspath

import sys

class HLAnalysis(Analysis):
    @property
    def name(self) -> str:
        return f'Highlight Analysis'
    def choose(self) -> bool:
        return True

    def __init__(cls):
        pass

    def run(cls):
        config = cls.config
        
        rest = HLRestCall(hl_base_url=config.hl_url,hl_user=config.hl_user,hl_pswd=config.hl_password,hl_instance=config.hl_instance)

        try:
            process = {}
            for appl in config.applist:
                app_name = appl['name']

                hl_status = appl['status']['highlight']=Status.ANALYSIS_START


                if hl_status == '' or hl_status.startswith('Error'):

                    # cls.log.info(f'Running Highlight analysis for {config.project_name}\{appl}')
                    cls.log.info(f'Running Highlight analysis for {appl}')
                    app_id = rest.get_app_id(appl)
                    if app_id is None:
                        cls.log.info(f'Application {appl} not found in Highlight, Creating Application {appl}...')
                        resp_code = rest.create_an_app(config.hl_instance, appl)
                        if resp_code == 200:
                            cls.log.info(f'Application {appl} created inside Highlight.')
                            app_id = rest.get_app_id(appl)
                        else:
                            cls.log.info(f'Not able to Application {appl} inside Highlight due to some error!')                       

                    hl_sourceDir = abspath(f'{config.work}\\{config.project_name}\\{appl}')
                    hl_workingDir = abspath(f'{config.oneclick_work}\\{config.project_name}\\HL_ANALYSIS_RESULT\\{appl}')
                    create_folder(f'{config.oneclick_work}\\{config.project_name}\\HL_ANALYSIS_RESULT')
                    create_folder(hl_workingDir)
                    java_home = config.java_home
                    if len(java_home) > 0:
                        java_home = f'{java_home}'

                    # args = [abspath(f'{config.java_home}/bin/java.exe'),
                    #         '-jar',config.hl_cli,
                    #         '--sourceDir', hl_sourceDir,
                    #         '--workingDir' , hl_workingDir,
                    #         '--companyId', str(config.hl_instance),
                    #         '--analyzerDir',config.analyzer_dir,
                    #         '--perlInstallDir',config.perl_install_dir,
                    #         '--applicationId', str(app_id),
                    #         '--serverUrl', config.hl_url.rsplit('/',1)[0],
                    #         '--login',config.hl_user,
                    #         '--password',config.hl_password]

                    args = [abspath(f'{config.base}/scripts/runHighlight.bat'),
                            abspath(f'{config.java_home}/bin/java.exe'),
                            config.hl_cli,
                            hl_sourceDir,
                            hl_workingDir,
                            str(config.hl_instance),
                            config.analyzer_dir,
                            config.perl_install_dir,
                            str(app_id),
                            config.hl_url.rsplit('/',1)[0],
                            config.hl_user,
                            config.hl_password]
                    try:
                        proc = run_process(args,wait=False)
                        pass

                    except FileNotFoundError as e:
                        cls.log.error(f'Unable to launch analysis process {e}')
                        cls.log.error(args)
                        return e.errno
                else:
                    proc = None

                cls.track_process(proc,"Highlight",appl)
            return 0


            # error = False
            # for appl in process:
            #     cls.log.info(f'Tracking Highlight analysis for {config.project_name}\{appl}')
            #     status,output = check_process(process[appl])
            #     if status != 0:
            #         cls.log.error(f'Error analyzing {appl}')
            #         error = True
            # if error:
            #     # TODO: add more desriptive error message
            #     raise RuntimeError ("")
        except KeyError as e:
            msg = str(e)
            if 'application not found:' in msg:
                cls.log.error(msg)
                sys.exit(1)


from os import walk,remove,stat
from os.path import abspath,exists

#from shutil import unpack_archive
#from pyunpack import Archive,ZipFile,error
from zipfile import ZipFile,BadZipFile
from gzip import open as gz
from tarfile import open as tar, TarError
from shutil import copyfileobj
from tqdm.auto import tqdm
from cast_common.util import create_folder
from cast_common.logger import Logger
from datetime import datetime

from oneclick.discovery.sourceValidation import SourceValidation 
from oneclick.configTest import Config,Status,App


class Unzip(SourceValidation):
   skip = ['__MACOSX','.DS_Store']

   @property
   def name(self) -> str:
      return __class__.__name__

   # def get_title(self) -> str:
   #    return 'DISCOVER ARCHIVES AND UNZIP'                               

   def __init__(self):
       pass

   def unzip(self,src_fldr:str,appl) -> tuple[bool,int,int]:
      unarchived_files = 0
      archived_files = 0
      error = False
      found = True

      loop = 0
      while found:
         loop +=1
         self.zip_log.info(f'Pass {loop}')
         
         found = False
         for root, dirs, files in walk(src_fldr):
            try:
               if '__MACOSX' in root or '.DS_Store' in files:
                  continue
               
               print (self.show_progress())

               for file in files:
                  full_name = abspath(f'{root.strip()}\\{file}')
                  dest = full_name.replace(f".{full_name.split('.')[-1]}",'')

                  file_type=''
                  if file.endswith(".gz"):
                     archived_files += 1
                     file_type = 'Gnu Zip'
                     with gz(full_name, 'rb') as f_in:
                        if not exists(dest):
                           with open(dest, 'wb') as f_out:
                              copyfileobj(f_in, f_out)               

                  elif file.endswith(".tar") or \
                       file.endswith(".gztar") or \
                       file.endswith(".bztar") or \
                       file.endswith(".tgz"):
                     with tar(full_name) as f_in:
                        archived_files += 1
                        file_type = 'Tape Archive (tar)'
                        create_folder(dest)
                        f_in.extractall(dest)

                  elif file.endswith(".7z") or file.endswith(".zip") :
                     archived_files += 1
                     file_type = 'zip archive'
                     full_name = abspath(f'{root.strip()}\\{file}')
                     create_folder(dest)
                     with ZipFile(full_name) as zf:
                        for member in tqdm(zf.infolist(), desc=f'Extracting {full_name}',position=0,leave=True):
                           zf.extract(member, dest)

                  else:
                     unarchived_files += 1

                  print (self.show_progress())

                  if len(file_type):    
                     found=True
                     appl['unpacked']+=1
                     file_stats = stat(full_name)
                     self.zip_log.info(f'Unpacked({file_type}): {full_name} {int(round(file_stats.st_size/1024,0))} KB')  
                     # self.log.info(f'Unpacked({file_type}): {full_name}')
                     remove(full_name)
                  else:
                     self.zip_log.info(f'Skipped {full_name}')
            except BadZipFile:
               #not a valid zip file log it then remove from staging folder
               print('')
               self.log.error(f'Bad zip file: {full_name}')  
               remove(full_name)

            except (TarError, ValueError, Exception) as ex:
               print('')
               self.log.error(f'Error while unpacking: {full_name}')  
               self.log.error(str(ex))
               error = True

      #terminate on error 
      if error: 
         raise
      return archived_files, unarchived_files

   # process_cnt = 0
   # process_txt = '\\|/-\\|/-'
   # def show_progress(self,done=False):
   #    config = self.config

   #    out = f'\r{self.process_txt[self.process_cnt]} Running {self.name} for project: {config.project_name}                                                       \n'
   #    out = out + 'Appl Name                                         Status                     Unpacked\n'
   #    out = out + '------------------------------------------------- ------------------------ ----------\n'

   #    for app in config.applist:
   #       app_name = app['name']
   #       app_status = app['status']['aip']
   #       if 'unpacked' not in app:
   #          app['unpacked'] = 0
   #       out = out + f"{app_name:<50}{Status(app_status).name.replace('_', ' '):<25}{app['unpacked']:>10}\n"

   #    if not done:
   #       for l in range(len(config.applist)+4):
   #          out = out + '\033[F'
   #       out = out + '\r'
   #       self.process_cnt += 1
   #       if self.process_cnt == len(self.process_txt):
   #             self.process_cnt = 0
   #       out = out + '\r\033[?25l'
   #    else:
   #       out = out + '\033[?25h'
   #       pass

   #    pass
   #    return out

   def run(self):
      config = self.config
      log = self.log

      log.debug('Unpack files')
      work_folder=abspath(f'{config.stage_folder}/{config.project_name}')


      for app in config.applist:
         app_status = app['status']
         if app_status['aip'] < Status.UNPACK_START:
            app['unpacked']=0

      print (self.show_progress())

      for app in config.applist:
         app_name = app['name']
         app_status = app['status']
         if app_status['aip'] < Status.STAGED:
            config.log.error(f'{app_name} must be staged before unziping files')
            return False
         elif app_status['aip'] > Status.UNPACK_END:
            config.log.debug(f'{app_name} already unpacked')
            continue

         app_name = app['name']
         app_status['highlight'] = app_status['aip']  = Status.UNPACK_START
         config._save()

         #configure logging for each app
         log_folder=abspath(f'{config.log_folder}/{config.project_name}/{app_name}/unpack')
         create_folder(log_folder)

         dateTimeObj=datetime.now()
         file_suffix=dateTimeObj.strftime("%d-%b-%Y(%H.%M.%S.%f)")

         zip_log_file= abspath(f"{log_folder}/{file_suffix}.log")
         log_name = 'Unzip-files'
         self.zip_log = Logger(log_name,level=config.log_level,file_name=zip_log_file,console_output=False)
         self.zip_log.info(f'{config.project}/{app}')

         archived_files, unarchived_files = self.unzip(abspath(f'{work_folder}/{app_name}'),app)
         print (self.show_progress())

         app['status']['highlight'] = app['status']['aip']  = Status.UNPACK_END
         config._save()

         msg = f'{config.project_name}/{app_name} found and unpacked {archived_files} archives found in {unarchived_files} files'
         # self.log.info(msg)
         self.zip_log.info(msg)
         # embedded_archives = archived_files - 1
         # if embedded_archives > 0:   
         #    self.log.info(f'{archived_files} files unzipped successfully. {embedded_archives} embedded archives found and unzipped successfully.\n')
         # else:
         #    self.log.info(f'{archived_files} files unzipped successfully. 0 embedded archives found.\n')
         Logger.loggers.remove(log_name)

      self.log.info(self.show_progress(True))
      self.log.debug('Archive file unpack complete')
      return True








from cast_common.logger import Logger,DEBUG, INFO, WARN, ERROR
from oneclick.config import Config
from cast_arg.config import Config as ARGConfig
from cast_arg.convert import GeneratePPT 
from cast_common.util import create_folder
from json import dump
from os.path import abspath

from subprocess import call

class RunARG():
    def __init__(cls,config:Config, log_level:int=INFO,log_name:str=None):
        if log_name is None:
            log_name = cls.__class__.__name__
            
        cls._log = Logger(log_name,log_level)
        cls._config = config
        pass

    def run(cls,config:Config,appl_name:str=None,operation:str=None) -> bool:
        if not config.is_aip_active:
            cls._log.warning('Dashboard configureaton is incomplete, Assessment report will NOT be generated')
            return -1

        if appl_name is None:
            appl_name = 'FULL'
        
        cfg = {}
        cfg['company']=config.company_name
        cfg['project']=config.project['name']
        cfg['template']=abspath(config.arg_template)
        cfg['output']=abspath(f'{config.report}/{config.project_name}/{appl_name}')
        cfg['cause']=abspath(f'{config.base}/scripts/cause.json')

        application=[]
        if appl_name == 'FULL':
            for appl in config.application:
                application.append({"aip":appl,"highlight":appl,"title":appl})
        else: 
            application.append({"aip":appl_name,"highlight":appl_name,"title":appl_name})
        cfg['application']=application

        cfg['rest'] = config.rest

        create_folder(cfg['output'])

        arg_cfg_file = abspath(f'{config.report}/{config.project_name}/{appl_name}-config.json')
        with open(arg_cfg_file, "w") as f:
            dump(cfg,f,indent=4)

        GeneratePPT(ARGConfig(arg_cfg_file)).save_ppt()
        pass

    def get_title(cls) -> str:
        return cls.__class__.__name__

class  RunARGAIP(RunARG):
    def __init__(cls, config:Config, log_level:int) -> None:
        super().__init__(config,log_level,cls.__class__.__name__)

    def run(cls,appl_name:str) -> None:
        super().run(cls._config,appl_name,'AIP')




__author__ = "Nevin Kaplan"
__copyright__ = "Copyright 2024, CAST Software"
__email__ = "n.kaplan@castsoftware.com"

from cast_common.logger import Logger,DEBUG, INFO, WARN, ERROR
from argparse import ArgumentParser
from os import listdir, getenv,getcwd
from os.path import abspath,exists,isfile, join
from argparse_formatter import FlexiFormatter,ParagraphFormatter
from json import load,dump

from inquirer import List,prompt,Text,password

from cast_common.util import create_folder, yes_no_input,folder_input,file_input,secret_input,url_input,string_input
from oneclick.exceptions import InvalidConfiguration,InvalidConfigNoBase
from enum import Enum,IntEnum

class Status(IntEnum):
    NOT_STARTED =                       0
    STAGED =                            10
    UNPACK_START =                      20
    UNPACK_END =                        25

    CLOC_PRE_CLEAN_START =              30
    CLOC_PRE_CLEAN_QUEUE =              32
    CLOC_PRE_CLEAN_RUNNING =            34
    CLOC_PRE_CLEAN_REPORT =             35
    CLOC_PRE_CLEAN_END =                36
    CLOC_PRE_CLEAN_ERROR =              38

    SOURCE_CLEAN_START =                40
    SOURCE_CLEAN_END =                  48

    CLOC_POST_CLEAN_START =             50
    CLOC_POST_CLEAN_QUEUE =             52
    CLOC_POST_CLEAN_REPORT =            35
    CLOC_POST_CLEAN_RUNNING =           54
    CLOC_POST_CLEAN_END =               56
    CLOC_POST_CLEAN_ERROR =             58

    SQL_DISCOVERY_START =               60
    SQL_DISCOVERY_END =                 68

    DISCOVERY_REPORT_START =            70
    DISCOVERY_REPORT_END =              78
    
    ANALYSIS_START =                    90
    ANALYSIS_QUEUE =                    92 
    ANALYSIS_END =                      95
    COMPLETED =                         100
    ERROR =                             999

    def __str__(self):    
        return str(self.value)

class Config():
    _config = None
    _config_file = None

    _base = None
    _log_base = None

    _log = None
    _log_level = None


    def command_line(self) -> ArgumentParser:
        parser = ArgumentParser(prog='OneClick',  formatter_class=lambda prog: FlexiFormatter(prog, width=99999, max_help_position=60))

        parser.add_argument('-b','--base', required=False, help='Base Folder Location',metavar='BASE_FOLDER')
        parser.add_argument('-p','--project', help='Name of the project')
        parser.add_argument('-s','--start', help='Starting step')
        parser.add_argument('-e','--end', help='Ending step')
        parser.add_argument('-g','--global_config',action='store_true',help='Update global configuration file')
        parser.add_argument('-d','--debug',action='store_true',help='Run in debug mode')
        parser.add_argument('-r','--reset',action='store_true',help='Rerun all applications')
        parser.add_argument('-q','--quiet',action='store_true',help='Run with no unnessary user intaction')

        parser.add_argument('--skipHighlight',action='store_true',help='Don\'t run Highlight Analysis')
        parser.add_argument('--skipMRI',action='store_true',help='Don\'t run MRI Analysis')

        # parser.add_argument('-rh','--runHighlight',action='store_true',default=True help='Run Highlight Analysis')
        # parser.add_argument('-rm','--runMRI',action='store_true',default=True help='Run MRI Analysis')

        return parser



#    def __init__(self,parser:ArgumentParser,default_args={},log_level: int=INFO):
    def __init__(self):

        Config._config = {}
        parser = self.command_line()
        args = parser.parse_args()

        Config._args = args
        base_config_name=None
        if not args.base is None:
            self.base = args.base
        if len(self.base)==0 or self.base is None:
            self.base = getcwd()
            # if not self.quiet:
            #     self.base = folder_input('Where is OneClick installed',folder=self.base,file='oneclick.bat')
            #     if not exists(self.base):
            #         raise InvalidConfiguration('Base configuration file found, please run setup')
            #make sure the configuration folder is there
                

        # load the default configuration, if exists
        base_config_name = abspath(f'{self.base}/.oneclick/config.json'.lstrip().rstrip())
        create_folder('\\'.join(base_config_name.split('\\')[:-1]))
        if exists(base_config_name):
            save_base=self.base
            with open(base_config_name, 'rb') as config_file:
                Config._config = load(config_file)
            self.base=save_base
            if args.global_config or \
               not self.check_list(self.settings,['work','java','cloc','profiler']) or \
               not self.check_list(self.highlight,['cli','agent']) or \
               not self.check_list(self.console, ['cli','URL', 'user', 'token']) :
                self.setup_default_config(False)
        else:
            self.quiet=False #global configuration file does not exist, we need to be sure quiet mode is not on
            self.setup_default_config()
            pass

        Config._log_level = DEBUG if args.debug else INFO
        Config.base_log_folder = abspath(f'{self.base}/logs')
        create_folder(self.base_log_folder)
        Config.log = Logger(self.__class__.__name__,DEBUG if args.debug else INFO,file_name=abspath(f'{self.base_log_folder}/General.log'))
        self.log.debug(self.report())

        #get the project name
        project_name = Config._config_file = None
        if not args.project is None:
            project_name = args.project.lstrip().rstrip()
        if project_name is None or len(project_name) == 0:

            dir_ = abspath(f'{self.base}/.oneclick')
            proj_cnt = len([name for name in listdir(dir_) if isfile(join(dir_, name))])
            if proj_cnt > 1:
                choices=[name.replace('.json','') for name in listdir(dir_) if isfile(join(dir_, name)) and name != 'config.json']
                choices.insert(0,'New Project')
                questions = [
                    List(
                        'project',
                        message="Which project do you want to work on?", 
                        choices=choices
                        )
                ]
                answers = prompt(questions)
                project_name = answers['project']
                if project_name == 'New Project':
                    questions = [
                        Text('project', message="What is the name of this project?")
                    ]
                    while True:
                        answers = prompt(questions)
                        project_name = answers['project']
                        project_config = abspath(f'{dir_}/{project_name}.json')
                        if not exists(project_config):
                            break
                        else:
                            if yes_no_input(f'Project {project_name} already exists, run it',default_value=False):
                                break
                            pass
                Config._config_file = abspath(f'{self.base}/.oneclick/{project_name}.json')
                pass
            else:
                project_name = string_input('What is the name of the project')
        
        Config._config_file = abspath(f'{self.base}/.oneclick/{project_name}.json')
        if exists(Config._config_file):
            with open(Config._config_file, 'rb') as config_file:
                Config._config = load(config_file)
        else:
            self.project_name = project_name

        self.merge_from_global()
        if not self.quiet and \
           not self.check_list(self.highlight,['URL','user','password','instance']) and \
           yes_no_input('Upload Highlight analysis results to the dashboard',default_value=True): 
            self.get_highlight_access_info()
        pass

    def merge_from_global(self):
        global_config_name = abspath(f'{self.base}/.oneclick/config.json'.lstrip().rstrip())
        global_data = None
        with open(global_config_name, 'rb') as config_file:
            global_data = load(config_file)

        self.copy_data(global_data['settings'], self.data['settings'])
        self.copy_data(global_data['rest']['highlight'], self.data['rest']['highlight'])
        self.copy_data(global_data['rest']['console'], self.data['rest']['console'])
        self.copy_data(global_data['rest']['dashboard'], self.data['rest']['dashboard'])

    def copy_data(self, gdata,pdata):
        """
        copy from gdata to pdata where gdata has a value and a change has been made to gdata

        Args:
            gdata (_type_): source dictionary 
            pdata (_type_): destination dictionary 
        """
        global_keys = set(gdata.keys())
        project_keys = set(pdata.keys())
        shared_keys = global_keys.intersection(project_keys)

        modified = False
        for key in shared_keys:
            if gdata is not None and len(gdata[key]) and pdata[key] != gdata[key]:
                pdata[key] = gdata[key]
                modified = True

        if modified:
            self._save()    
        pass


    def setup_default_config(self,first=True):
        if not self.update_global and self.quiet:
            return 
        
        msg = ''
        if first:
            msg = 'This is the first time you are running OneClick, we need to set up some default configuration.'
        else:
            msg = 'Detected some missing configuration information, rerunning setup now.'
        print (
"""
Global Configuration
--------------------
{msg}

OneClick is organized into two distinct components: the base and the work folders. The base 
encompasses all executable files and configuration data, while the work folder houses the 
source code to be analyzed, as well as all logging, analysis results and reports. Although this 
division is not mandatory, it facilitates usage over a network connection. Notably, the base 
and work folders can also reside in the same location.

""".format(msg=msg)
        )

        Config._config_file = abspath(f'{self.base}/.oneClick/config.json')
        # if exists(self._config_file):
        #     with open(self._config_file, 'rb') as config_file:
        #         Config._config = load(config_file)

        scripts_folder = abspath(f'{self.base}/scripts')

        #work folder
        if self.update_global or not self.quiet:
            self.work = folder_input('Work folder location',folder=self.work,create=True) 
            create_folder(self.deliver_folder)
            create_folder(self.stage_folder)
            create_folder(self.report_folder)
            create_folder(self.highlight_folder)

            #java home
            self.java_home = folder_input("Java installation folder location",getenv('JAVA_HOME'),'bin/java.exe')
            #cloc executable location
            self.cloc = file_input("CLOC installation folder location", scripts_folder, 'cloc*.exe')
            #profiler executable location
            self.profiler = folder_input("Profiler installation folder location",self.profiler,'CAST-Profiler.exe')

            print ('')

        #highlight rest api information
        if first or not (self.highlight and self.check_list(self.highlight,['cli','agent','perl','URL','user','password','instance'])):
            if yes_no_input('Is OneClick being used to run Highlight analysis?'):
                self.hl_cli = folder_input('Highlight Automation Command location',self.hl_cli,'HighlightAutomation.jar',True)
                self.hl_agent = folder_input('Highlight Code Reader location',self.hl_agent,'nw.exe',True)
                agent_no_file = self.hl_agent.rsplit("\\",1)[0]
                # self.hl_perl = abspath(f'{agent_no_file}/strawberry')
                # self.hl_analysis = abspath(f'{self.hl_cli}/strawberry/perl')

                if yes_no_input('Upload results to highlight dashboard',default_value=True):
                    self.get_highlight_access_info()
                    pass
                print ('')
        
        #console automation information
        if first or not (self.console and self.check_list(self.console,['cli','URL','user','token'])):
            if yes_no_input('Is OneClick being used to run MRI analysis?'):
                self.console_cli = folder_input('MRI Automation Command Line interface location', self.console_cli, 'aip-console-tools-cli.jar', True)
                self.console_url = url_input('AIP Console URL', self.console_url)
                self.console_user = string_input('AIP Console User ID', self.console_user)
                self.console_token = secret_input('AIP Console Token',self.console_token)
                print ('')
            else:
                self._save()
        pass


        """
        collect dashboard information
            check if the console information is there, if not no MRI analysis 
            will be run so there is no need to add dashboard
        """
        if self.console and first and not (self.quiet and self.dashboard):
            questions = [
                Text('url', message="What is the URL of the CAST Dashboard"),
                Text('user','What is the user id'),
                password('user','What is the password'),
            ]
            answers = prompt(questions)
            self.dashboard_url = answers['url']
            self.dashboard_user = answers['user']
            self.dashboard_password = answers['password']   
            pass
        else:
            self.dashboard
            self._save()

        pass

    def get_highlight_access_info(self):
        self.hl_url = url_input('Highlight URL',self.hl_url)
        self.hl_user = string_input('Highlight user name',self.hl_user)
        self.hl_password = secret_input('Highlight password')
        self.hl_instance = string_input('Highlight instance id',self.hl_instance)
        pass

    def report(self,is_config=False):
        """
        Report the current configuration
        """
        rpt_base =  ''
        if self.debug:
            rpt_base = rpt_base + \
"""
     ***********************************************************************************************
                          R U N N I N G  I N   D E B U G   M O D E
     ***********************************************************************************************
"""            

        rpt_base = rpt_base +  \
"""
    OneClick Configuration
    ----------------------
    Base Folder: {self.base}
    General Settings:
        Work Folder: {self.work}
        Java Home: {self.java_home}
        CLOC: {self.cloc}
        Profiler: {self.profiler}
    REST Configuration
        Highlight:
            CLI: {self.hl_cli}
            Agent: {self.hl_agent}
            URL: {self.hl_url}
            User: {self.hl_user}
            Instance: {self.hl_instance}
            password: {hl_password}
        Console:
            AIP Console CLI: {self.console_cli}
            URL: {self.console_url}
            User: {self.console_user}
            Token: {console_token}
        Dashboard:
            URL: {self.dashboard_url}
            User: {self.dashboard_user}
            Password: {dashboard_password}
    """.format(self=self,
               hl_password=('Password Set' if len(self.hl_password) else 'Password Not Set'),
               console_token=('Password Set' if len(self.console_token) else 'Password Not Set'),
               dashboard_password=('Password Set' if len(self.dashboard_password) else 'Password Not Set')
        )

        if is_config:
            return rpt_base
        rpt_apps = self.application_report()

        rpt_final = """
========================================================================================================================
    """
        rpt_txt = rpt_base + rpt_apps + rpt_final
        return rpt_txt

    def application_report(self):
        rpt_apps = f" Project: {self.project_name}"
        if len(self.applist) == 0:
            rpt_apps = rpt_apps + "\tNo Applications"
        else:
            rpt_apps = rpt_apps + '\n Appl Name\tHighlight\t                 MRI\t\t\t\tMRI Security    MRI Blue Print\n'
            rpt_apps = rpt_apps + ' ----------\t-------------------------------\t------------------------------- --------------- ------------------\n'
            for app in self.applist:
                # app['status']['aip']=app['status']['highlight']=Status.NOT_STARTED
                rpt_apps = rpt_apps + f" {app['name']:<15}\t{Status(app['status']['highlight']).name.replace('_',' '):<25}\t{Status(app['status']['aip']).name.replace('_',' '):<25}\t{app['security']}\t\t{app['blueprint']}\n"
        return rpt_apps        

    def _get(self,base,key,default=''):
        if key not in base:
            base[key] = default  
        return base[key]        

    def _set_value(self,base,key,value,default=''):
        base[key] = self._get(base,key,default)
        if value is not None:
            base[key]=value
            return True
        else:
            return False

    def check_list(self, node, chk_lst):
        chk = True
        for item in chk_lst:
            if item not in node or node[item]=='':
                chk = False
                break
        return chk
    
    def _set_active(self,node,chk_lst):
        node['Active']=self.check_list(node, chk_lst)

    def _save(self):
        if Config._config_file is None: raise InvalidConfiguration('No config file')
        create_folder(abspath(f'{self.base}/.oneClick'))
        with open(Config._config_file, "w") as f:
            dump(Config._config, f,indent=4)

    @property
    def data(self):
        return Config._config

    """
        Argument Properties
    """
    @property
    def quiet(self):
        return Config._args.quiet 
    @quiet.setter
    def quiet(self, value):
        Config._args.quiet = value
    @property
    def start(self):
        return Config._args.start 
    @start.setter
    def start(self, value):
        Config._args.start = value

    @property
    def end(self):
        return Config._args.end 
    @end.setter
    def end(self, value):
        Config._args.end = value

    @property
    def debug(self):
        return Config._args.debug
    @property
    def update_global(self):
        return Config._args.global_config

    @property
    def reset(self):
        return Config._args.reset 


    @property 
    def deliver_folder(self): return abspath(f'{self.work}/DELIVER')
    @property 
    def stage_folder(self): return abspath(f'{self.work}/STAGED')
    @property 
    def report_folder(self): return abspath(f'{self.work}/REPORT')
    @property 
    def highlight_folder(self): return abspath(f'{self.work}/HIGHLIGHT')
    @property 
    def log_folder(self): return abspath(f'{self.base}/LOGS')
    @property 
    def scripts_folder(self): return abspath(f'{self.base}/SCRIPTS')

    @property
    def base(self):
        return self._get(Config._config,'base')
    @base.setter
    def base(self,value):
        self._set_value(Config._config,'base',value)
        
    @property
    def base_log_folder(self):
        return self._get(Config._config,'base.log.folder')
    @base_log_folder.setter
    def base_log_folder(self, value):
        self._set_value(Config._config, 'base.log.folder', value)

    @property
    def work(self):
        return self._get(self.settings,'work')
    @work.setter
    def work(self, value):
        self._set_value(self.settings, 'work', value)
        self._save()

    @property
    def log(self) -> Logger:
        return Config._log
    @property
    def log_level(self):
        return Config._log_level
    
    @property
    def settings(self):
       return self._get(Config._config,'settings',{})

    @property
    def java_home(self) -> str:
        return self._get(self.settings,'java','')
    @java_home.setter
    def java_home(self,value):
        self._set_value(self.settings, 'java', value)
        self._save()

    @property
    def cloc(self) -> str:
        return self._get(self.settings,'cloc','')
    @cloc.setter
    def cloc(self, value):
        self._set_value(self.settings, 'cloc', value)
        self._save()

    @property
    def profiler(self) -> str:
        return self._get(self.settings, 'profiler', '')
    @profiler.setter
    def profiler(self, value):
        self._set_value(self.settings, 'profiler', value)
        self._save()

    @property
    def rest(self):
       return self._get(Config._config,'rest',{})

    @property
    def highlight(self):
       return self._get(self.rest,'highlight',{})

    @property
    def hl_cli(self) -> str:
        return self._get(self.highlight,'cli','')
    @hl_cli.setter
    def hl_cli(self, value):
        self._set_value(self.highlight, 'cli', value)
        # self._set_hl_active()
        self._save()
        
    @property
    def hl_agent(self) -> str:
        return self._get(self.highlight,'agent','')
    @hl_agent.setter
    def hl_agent(self, value):
        self._set_value(self.highlight, 'agent', value)
        self._save()
    # @property
    # def hl_analizer_folder(self) -> str:
    #     agent = abspath(self._get(self.highlight, 'agent', ''))
    #     agent = agent.rsplit("\\", 1)[0]
    #     return abspath(f'{agent}/perl')

    # @property
    # def hl_perl(self) -> str:
    #     return abspath(f"{self._get(self.highlight,'perl','')}/perl")
    # @hl_perl.setter
    # def hl_perl(self, value):
    #     self._set_value(self.highlight, 'perl', value)
    #     self._save()

    @property
    def hl_user(self) -> str:
        return self._get(self.highlight, 'user', '')
    @hl_user.setter
    def hl_user(self, value):
        self._set_value(self.highlight, 'user', value)
        self._save()

    @property
    def hl_url(self) -> str:
        return self._get(self.highlight, 'URL', '')
    @hl_url.setter
    def hl_url(self, value):
        self._set_value(self.highlight, 'URL', value)
        self._save()

    @property
    def hl_password(self) -> str:
        return self._get(self.highlight, 'password', '')
    @hl_password.setter
    def hl_password(self, value):
        self._set_value(self.highlight, 'password', value)
        self._save()

    @property
    def highlight_active(self):
        self.check_list(self.highlight,['hl_cli','hl_agent','hl_perl','URL','hl_user','hl_instance'])

    @property
    def hl_instance(self) -> str:
        return self._get(self.highlight, 'instance', '')
    @hl_instance.setter
    def hl_instance(self, value):
        self._set_value(self.highlight, 'instance', value)
        self._save()

    @property
    def console(self):
       return self._get(self.rest, 'console', {})

    @property
    def console_cli(self) -> str:
        return self._get(self.console, 'cli', '')
    @console_cli.setter
    def console_cli(self, value):
        self._set_value(self.console, 'cli', value)
        self._save()

    @property
    def console_url(self) -> str:
        return self._get(self.console, 'URL', '')
    @console_url.setter
    def console_url(self, value):
        self._set_value(self.console, 'URL', value)
        self._save()

    @property
    def console_token(self) -> str:
        return self._get(self.console, 'token', '')
    @console_token.setter
    def console_token(self, value):
        self._set_value(self.console, 'token', value)
        self._save()

    @property
    def console_user(self) -> str:
        return self._get(self.console, 'user', '')  
    @console_user.setter
    def console_user(self, value):
        self._set_value(self.console, 'user', value)
        self._save()

    @property
    def dashboard(self):
       return self._get(self.rest, 'dashboard', {})

    @property
    def dashboard_url(self) -> str:
        return self._get(self.dashboard, 'URL', '')
    @dashboard_url.setter
    def dashboard_url(self, value):
        self._set_value(self.dashboard, 'URL', value)
        self._save()

    @property
    def dashboard_user(self) -> str:
        return self._get(self.dashboard, 'user', '')
    @dashboard_url.setter
    def dashboard_user(self, value):
        self._set_value(self.dashboard, 'user', value)
        self._save()

    @property
    def dashboard_password(self) -> str:
        return self._get(self.dashboard, 'password', '')
    @dashboard_url.setter
    def dashboard_password(self, value):
        self._set_value(self.dashboard, 'password', value)
        self._save()

    @property
    def enable_security_assessment(self) -> bool:
        return self._get(self.console, 'enable_security_assessment', True)
    @enable_security_assessment.setter
    def enable_security_assessment(self, value):
        self._set_value(self.console, 'enable_security_assessment', value)
        self._save()

    @property
    def project(self):
       return self._get(Config._config, 'project', {})
    
    @property
    def project_name(self):
       return self._get(self.project,'name')
    @project_name.setter
    def project_name(self, value):
        self._set_value(self.project,'name', value)
        self._save()

    @property
    def is_hl_active(self) -> bool:
        return self.highlight['Active']
    
    @property
    def applist(self) -> List:
       return self._get(self.project, 'applist', [])

    def add_app(self, app_name,security,blueprint):
        if app_name not in self.applist:
            new_app = App(app_name)
            new_app.security = security
            new_app.blueprint = blueprint
            self.applist.append(new_app)
            # self._set_value(self.applist, app_name, new_app)
            self._save()  # save the config after adding an app
            return new_app
        else:
            return self._applist[app_name]

    def remove_app(self, app_name):
        if app_name in self._applist:
            del self._applist[app_name]
            self._save()  
            return True
        return False

class App(dict):
    def __init__(self, name):
        self.name = name
        self['status'] = {'highlight': Status.NOT_STARTED, 'aip': Status.NOT_STARTED}

    @property
    def name(self):
        if 'name' not in self: return None
        return self['name']
    @name.setter
    def name(self, value):
        self['name'] = value

    @property
    def status(self):
        return self['status']

    @property
    def highlight(self):
        return self.status['highlight']
    @highlight.setter
    def highlight(self, value:Status):
        if not isinstance(value, Status):
            raise ValueError("Status must be a Status enum")
        self.status['highlight'] = value
        self.log.info(f'Highlight status set to {value.value}')

    @property
    def aip(self):
        return self.status['aip']
    @aip.setter
    def aip(self, value:Status):
        if not isinstance(value, Status):
            raise ValueError("Status must be a Status enum")
        self.status['aip'] = value
        self.log.info(f'CAST Imaging status set to {value.value}')

    @property
    def is_security(self) ->bool:
        return self['security']
    @is_security.setter
    def security(self, value:bool):
        if type(value) is not bool: raise ValueError("Security must be a boolean")
        self['security'] = value
    
    @property
    def is_blueprint(self) -> bool:
        return self['blueprint']
    @is_blueprint.setter
    def blueprint(self, value: bool):
        if type(value) is not bool: raise ValueError("Blueprint must be a boolean")
        self['blueprint'] = value





from oneclick.configTest import Config,Status
from oneclick.analysis.analysis import Analysis
from oneclick.discovery.cleanup import Cleanup
from cast_common.hlRestCall import HLRestCall
from cast_common.util import run_process,create_folder,check_process
from subprocess import Popen,PIPE

from os.path import abspath
from os.path import exists

class RunAnalysis(Analysis):
    _running = 0
    _max = 2
    _process = {}
    _output = {}

    @property
    def max(cls):
        return RunAnalysis._max
    @property
    def running(cls):
        return RunAnalysis._running
    @running.setter
    def running(cls,value):
        RunAnalysis._running = value

    @property
    def process(cls):
        return RunAnalysis._process

    @property
    def output(cls) -> dict:
        return RunAnalysis._output

    def __init__(cls):
        # RunAnalysis._max = 2
        # RunAnalysis._running = 0
        pass

    def run(self):

        

        config = self.config
        while True:
            done = True
            for appl in config.applist:
                print (self.show_progress())
                app_name = appl['name']
                app_status = Status(self.status(appl))
                if app_status < Status.SOURCE_CLEAN_END:
                    config.log.info(f'Analysis cannot run, please run "{Cleanup.name}" step first')
                    continue
                if app_status >= Status.ANALYSIS_END:
                    continue
                done = False

                if self.can_run(appl):
                    self.log.info(f'Running {self.name} for {config.project_name}\{app_name}')

                    proc = self.run_analysis(appl)
                    if proc == 'DONE':
                        app_status = self.status(appl, Status.ANALYSIS_END)
                        pass
                    else:
                        app_status = self.status(appl, Status.ANALYSIS_START)
                        self.running += 1
                        self.process[app_name] = proc
                        self.output[app_name] = []
                        pass
                    print(self.show_progress())
                elif app_status == Status.ANALYSIS_START:
                    #get the process results
                    if (app_name in self.process):
                        if (self.process[app_name].poll() is None):
                            #process is still running, collect some output
                            out_ary = self.output[app_name]
                            line = self.process[app_name].stdout.readline()
                            if line and len(line) > 0:
                                out_ary.append(line)
                            pass
                        else:
                            #process is done, mark it complete and read any remaining lines 
                            app_status = self.status(appl, Status.ANALYSIS_END)
                            self.running -= 1
                            try:    
                                out_ary = self.output[app_name]
                                while True:
                                    line = self.process[app_name].stdout.readline()
                                    if line and len(line) > 0:
                                        out_ary.append(line)
                                    elif not line:
                                        break
                                
                            except IOError as ex:
                                if str(ex) == 'read of closed file':
                                    config.log.debug(f'Process already closed {app_name}')
                            pass
                else:
                    app_status = self.status(appl, Status.ANALYSIS_QUEUE)
                    pass
            if done:
                break

        print (self.show_progress(done=True))
        return True


    process_cnt = 0
    process_txt = '\\|/-\\|/-'
    def show_progress(self,done=False,clear=False):
        config = self.config
        out = []

        header = f'------------------------ ------------------------ ----------------------------------------------------------------------------------------------------'
        if clear:
            for l in range(len(config.applist)+5):
                line = ' '
                out.append(f'{line:{len(header)}}\n')
        else:
            line = f'{self.process_txt[self.process_cnt]} Running {self.name} for project: {config.project_name}'
            out.append(f'\r{line:<{len(header)}}\n')
            out.append(f'Appl Name                Status\n')
            out.append(f'{header}\n')
            for app in config.applist:
                app_status = self.status(app)
                app_name = app['name']
                last_line = self.output[app_name][-1].replace('\n','').replace('\t','') if app_name in self.output and len(self.output[app_name])>0 else ''
                last_line = last_line[:100].ljust(100)
                out.append(f'{app_name:<25}{Status(app_status).name.replace("_", " "):<25}{last_line}\n')

        if not done:
            for l in range(len(config.applist)+4):
                out.append(f'\033[F')
            out.append('\r')
            self.process_cnt += 1
            if self.process_cnt == len(self.process_txt):
                self.process_cnt = 0
            out.append(f'\r\033[?25l')
        return ''.join(out)

    def can_run(self, appl):
        app_status = self.status(appl)
        app_name = appl['name']

        if self.running >= self.max:
            # too many analysis running
            return False

        if app_status in [Status.ANALYSIS_START, Status.ANALYSIS_QUEUE] and app_name not in self.process:
            return True
        return False


class RunHighlight(RunAnalysis):

    def __init__(cls):
        # RunHighlight.__running = 0
        # RunHighlight.__max = 2
        
        cls._df = {}
        config = cls.config
        cls.rest = HLRestCall(hl_base_url=config.hl_url,hl_user=config.hl_user,hl_pswd=config.hl_password,hl_instance=config.hl_instance)
        pass

    @property
    def choose(cls) -> bool:
        return True
    @property
    def name(cls) -> str:
        return 'RunHighlight'


    # @property
    # def running(cls):
    #     return RunHighlight.__running
    # @running.setter
    # def running(value):
    #     RunHighlight.__running = value

    # @property
    # def max(cls):
    #     return RunHighlight.__max

    def status(cls,appl,new_status=None):
        if new_status is not None:
            appl['status']['highlight'] = new_status
            cls.config._save()
        return appl['status']['highlight']

    def run_analysis(cls, appl):
        config = cls.config
        app_name = appl['name']

        hl_source_folder = abspath(f'{config.stage_folder}/{config.project_name}\\{app_name}')
        hl_working_folder = abspath(f'{config.highlight_folder}/{config.project_name}/HL_ANALYSIS_RESULT/{app_name}')
        create_folder(hl_working_folder)
        java_home = config.java_home
        if len(java_home) > 0:
            java_home = f'{java_home}'

        do_upload=True
        url = config.hl_url.rsplit('/',1)[0]
        
        app_id = None
        if len(url) == 0 or len(config.hl_user) == 0 or len(config.hl_password) == 0:
            cls.log.info(f'Highlight URL not defined, skipping analysis upload for {app_name}')
            do_upload = False
        else:
            app_id = cls.rest.get_app_id(app_name)


        if app_id is None and do_upload:
            config.log.info(f'Application not found in Highlight, Creating Application {app_name}...')
            resp_code = cls.rest.create_an_app(config.hl_instance, app_name)
            if resp_code == 200:
                config.log.info(f'Application {app_name} successfuly created inside Highlight.')
                app_id = cls.rest.get_app_id(app_name)
            else:
                config.log.error(f'Error creating Application {app_name} inside Highlight due to some error!') 

        hl_cli_folder = abspath(config.hl_cli.rsplit('\\',1)[0])


        args = [abspath(f'{config.base}/scripts/runHighlight.bat'),
                abspath(f'{config.java_home}/bin/java.exe'),
                hl_cli_folder,
                hl_source_folder,
                hl_working_folder,
                config.hl_url,
                str(config.hl_instance),
                str(app_id),
                config.hl_user,
                config.hl_password]

        try:
            proc = run_process(args,wait=False)
            return proc
            pass

        except FileNotFoundError as e:
            cls.log.error(f'Unable to launch analysis process {e}')
            cls.log.error(args)
            return e.errno

        return 'DONE'



from cast_common.logger import Logger,INFO
from oneclick.config import Config
from cast_common.util import run_process
from subprocess import Popen
from oneclick.base import Base

class Analysis(Base):

    _pid = []

    def __init__(cls):
        pass

    def track_process(cls,process:Popen,operation:str,name:str):
        Analysis._pid.append(Process(process,operation,name))
        pass

    def check_process(cls,pid:int)->Popen:
        info = Analysis._pid[str(pid)]
        process = info['process']
        return process.poll()
    
    def get_title(cls) -> str:
        return cls.__class__.__name__

    

class Process():
    def __init__(cls,process:Popen,operation:str,name:str):
        cls._process = process
        cls._operation = operation
        cls._name = name
        cls._status = None
        cls._log = []

    @property
    def process(cls):
        return cls._process

    @property
    def operation(cls):
        return cls._operation

    @property
    def name(cls):
        return cls._name

    @property
    def status(cls):
        return cls._status
    @status.setter
    def status(cls,value):
        cls._status = value

    @property
    def log(cls):
        return cls._log



from oneclick.discovery.sourceValidation import SourceValidation 
from oneclick.config import Config
from cast_common.logger import Logger,INFO
from cast_common.util import run_process,track_process,check_process,format_table,create_folder
from os import listdir,remove,rename
from os.path import exists,abspath,getsize
from pandas import json_normalize,ExcelWriter
from json import load
from tqdm import tqdm



class ProfilerPreCleanup(SourceValidation):
    @property
    def name(self) -> str:
        return f'CAST Profiler'

    def __init__(cls):
        cls._df = {}
        pass

    # @property
    # def phase(cls):
    #     return 'Before'

    def _run_profiler(cls,app_name:str, work_folder:str,profiler_output:str):

        profiler_path=cls.config.profiler
        profiler_output=abspath(f'{cls.config.report}\{cls.config.project_name}\{app_name}')
        work_folder=abspath(f'{cls.config.work}/AIP/{cls.config.project_name}\{app_name}')

        create_folder(profiler_output)

        args = [profiler_path,
                work_folder,
                '--generate-report', profiler_output,
                '--name',app_name,
                '--output',profiler_output,
                '--offline','--details','--no-upload','--no-browser','--complete-insight'
            ]
        cls._log.info(' '.join(args))

        #cleanup any old files
        files = [f for f in listdir(profiler_output) if f.startswith(app_name) and f.endswith('result.json')]
        for f in files:
            remove (f'{profiler_output}/{f}')

        # run profiler
        proc = run_process(args,wait=False)
        track_process(proc)

        #rename the profile output file to a standard name
        files = [f for f in listdir(profiler_output) if f.startswith(app_name) and f.endswith('result.json')]
        new_name = abspath(f'{profiler_output}/{app_name}-{cls.phase}.json')
        if len(files) > 0:
            old_name = abspath(f'{profiler_output}/{files[0]}')
            if exists(new_name):
                remove(new_name)
            rename(old_name,new_name)

        #process results
        if exists(new_name):
            with open(new_name) as f:
                prflr = load(f)

            #process the data
            alerts = json_normalize(prflr['alerts'])

            composition = json_normalize(prflr['composition'],max_level=1)
            omposition = json_normalize(prflr['composition'],max_level=1)
            if 'subExtensions' in composition.columns:
                composition = composition.explode(column=['subExtensions']).fillna('')

            dependencies = json_normalize(prflr['dependencies'])
            if 'iconNames' in dependencies.columns:
                dependencies = dependencies.explode(column=['iconNames'])
            if 'versions' in dependencies.columns:
                dependencies = dependencies.explode(column=['versions'])

            frameworks = json_normalize(prflr['frameworks'])
            ext_list = json_normalize(prflr['extensions_list'],max_level=1).transpose().reset_index()
            ext_list = ext_list.rename(columns={'index':'Name',0:'Count'})
            files = json_normalize(prflr['files'],max_level=1)	

            # add results to excel sheet
            file_name = abspath(f'{profiler_output}/{app_name}-{cls.phase.lower()}-prfl-rslts.xlsx')
            writer = ExcelWriter(file_name, engine='xlsxwriter')
            if not alerts.empty: format_table(writer,alerts,'alerts',total_line=True)
            if not composition.empty: format_table(writer,composition,'composition',total_line=True)
            if not dependencies.empty: format_table(writer,dependencies,'dependencies',total_line=True)
            if not ext_list.empty: format_table(writer,ext_list,'ext_list',total_line=True)
            if not files.empty: format_table(writer,files,'files',total_line=True)
            if not frameworks.empty: format_table(writer,frameworks,'frameworks',total_line=True)
            writer.close()
        else: 
            cls._log.warning('No profiler results found')
        pass

    def run(cls):
        for appl in config.application:
            work_folder = abspath(f'{config.work}/AIP/{config.project_name}/{appl}')
            profiler_output=abspath(f'{config.report}/{config.project_name}/{appl}')
            
            create_folder(profiler_output)
            cls._run_profiler(app_name=appl,work_folder=work_folder,profiler_output=profiler_output)
            
